__all__ = ["is_small_molecule", "standardize"]


import biotite.structure as struc


def is_small_molecule(chain: struc.AtomArray) -> bool:
    """
    Check whether the given chain is a small molecule.

    Parameters
    ----------
    chain : struc.AtomArray, shape=(n,)
        The chain to check.

    Returns
    -------
    bool
        Whether the chain is a small molecule.
    """
    return chain.hetero[0].item()


def standardize(
    system: struc.AtomArray | struc.AtomArrayStack,
) -> struc.AtomArray | struc.AtomArrayStack:
    """
    Standardize the given system.

    This function

    - removes hydrogen atoms
    - removes solvent atoms and monoatomic ions

    Parameters
    ----------
    system : struc.AtomArray or struc.AtomArrayStack
        The system to standardize.

    Returns
    -------
    struc.AtomArray or struc.AtomArrayStack
        The standardize system.
    """
    mask = (
        (system.element != "H")
        & ~struc.filter_solvent(system)
        & ~struc.filter_monoatomic_ions(system)
    )
    return system[..., mask]



__all__ = ["GraphMatchWarning", "find_matching_atoms"]


import itertools
import warnings
from typing import Any
import biotite.interface.rdkit as rdkit_interface
import biotite.sequence as seq
import biotite.sequence.align as align
import biotite.structure as struc
import numpy as np
from numpy.typing import NDArray
from rdkit.Chem import (
    AssignStereochemistryFrom3D,
    BondType,
    Mol,
    SanitizeFlags,
    SanitizeMol,
)
from peppr.common import is_small_molecule

# To match atoms between pose and reference chain,
# the residue and atom name are sufficient to unambiguously identify an atom
_ANNOTATIONS_FOR_ATOM_MATCHING = ["res_name", "atom_name"]
_IDENTITY_MATRIX = align.SubstitutionMatrix(
    seq.ProteinSequence.alphabet,
    seq.ProteinSequence.alphabet,
    np.eye(len(seq.ProteinSequence.alphabet), dtype=np.int32),
)


class GraphMatchWarning(UserWarning):
    """
    This warning is raised, if the RDKit based molecule matching fails.
    In this case small molecule reordering is skipped.
    """

    pass


def find_matching_atoms(
    reference: struc.AtomArray,
    pose: struc.AtomArray,
    min_sequence_identity: float = 0.95,
) -> tuple[NDArray[np.int_], NDArray[np.int_]]:
    """
    Find the optimal atom order for each pose that minimizes the RMSD to the reference.

    Parameters
    ----------
    reference : struc.AtomArray, shape=(n,)
        The reference structure.
    pose : struc.AtomArray, shape=(n,)
        The pose structure.
    min_sequence_identity : float
        The minimum sequence identity between two chains to be considered the same
        entity.

    Returns
    -------
    reference_order, : np.array, shape=(n,), dtype=int
        The atom order that should be applied to `reference`.
    pose_order : np.array, shape=(n,), dtype=int
        The atom order that should be applied to `pose`.

    Notes
    -----
    For finding the optimal chain permutation, the function uses the
    algorithm described in [1]_.

    References
    ----------
    .. [1] *Protein complex prediction with AlphaFold-Multimer*, Section 7.3, https://doi.org/10.1101/2021.10.04.463034
    """
    ref_chain_starts = struc.get_chain_starts(reference, add_exclusive_stop=True)
    mod_chain_starts = struc.get_chain_starts(pose, add_exclusive_stop=True)
    reference_chains = [
        reference[start:stop] for start, stop in itertools.pairwise(ref_chain_starts)
    ]
    pose_chains = [
        pose[start:stop] for start, stop in itertools.pairwise(mod_chain_starts)
    ]

    # Find corresponding chains by identifying the chain permutation minimizing the RMSD
    mod_chain_order, transform, anchor_index = _find_matching_chain_permutation(
        reference_chains, pose_chains, min_sequence_identity
    )

    # Based on the matching chains, find corresponding atoms within each pair of chains
    ref_atom_orders: list[np.ndarray | None] = [None] * len(reference_chains)
    mod_atom_orders: list[np.ndarray | None] = [None] * len(reference_chains)
    for ref_i, mod_i in zip(np.arange(len(reference_chains)), mod_chain_order):
        if is_small_molecule(reference_chains[ref_i]):
            try:
                ref_atom_orders[ref_i], mod_atom_orders[mod_i] = (
                    _find_optimal_molecule_permutation(
                        reference_chains[ref_i],
                        transform.apply(pose_chains[mod_i]),
                        # If no other chain already determined the optimal transform,
                        # superimpose each permuted molecule to get the best RMSD
                        superimpose=True if mod_i == anchor_index else False,
                    )
                )
            except Exception as e:
                warnings.warn(
                    f"RDKit failed atom matching: {e}",
                    GraphMatchWarning,
                )
                ref_atom_orders[ref_i] = np.arange(
                    reference_chains[ref_i].array_length()
                )
                mod_atom_orders[mod_i] = np.arange(pose_chains[mod_i].array_length())
        else:
            ref_atom_orders[ref_i], mod_atom_orders[mod_i] = _find_common_residues(
                reference_chains[ref_i], pose_chains[mod_i]
            )

    # Finally bring chain order and order within chains together
    return (
        _combine_chain_and_atom_orders(
            np.arange(len(reference_chains)), ref_atom_orders, ref_chain_starts
        ),
        _combine_chain_and_atom_orders(
            mod_chain_order, mod_atom_orders, mod_chain_starts
        ),
    )


def _combine_chain_and_atom_orders(
    chain_order: NDArray[np.int_],
    atom_orders: list[NDArray[np.int_]],
    chain_starts: NDArray[np.int_],
) -> NDArray[np.int_]:
    """
    Bring both atom and chain orders together.

    Parameters
    ----------
    chain_order : np.ndarray, shape=(k,), dtype=int
        The order of the chains in terms of indices pointing to a list of chains.
    atom_orders : list of (np.ndarray, shape=(n,), dtype=int), length=k
        The order of the atoms in terms of indices pointing to atoms within each chain.
    chain_starts : np.ndarray, shape=(k+1,), dtype=int
        The start indices of the chains.
        Contains the exclusive stop index of the last chain as last element.

    Returns
    -------
    global_order : np.ndarray, shape=(n,), dtype=int
        The order of the atoms in the global system.
    """
    order_chunks = []
    # Take atom indices for each chain in the determined order
    for chain_i in chain_order:
        atom_indices = np.arange(chain_starts[chain_i], chain_starts[chain_i + 1])
        # Apply reordering within the chain
        atom_indices = atom_indices[atom_orders[chain_i]]
        order_chunks.append(atom_indices)
    return np.concatenate(order_chunks)


def _find_matching_chain_permutation(
    reference_chains: list[struc.AtomArray],
    pose_chains: list[struc.AtomArray],
    min_sequence_identity: float,
) -> tuple[NDArray[np.int_], struc.AffineTransformation, int]:
    """
    Find the permutation of the given chains that minimizes the RMSD between the pose
    and the reference.

    Parameters
    ----------
    reference_chains : list of struc.AtomArray, length=n
        The reference chains.
    pose_chains : list of struc.AtomArray, length=n
        The pose chains.
        Must have the same order as the `reference_chains`,
        i.e. elements at the same index must be equivalent chains.
    min_sequence_identity : float
        The minimum sequence identity between two chains to be considered the same
        entity.

    Returns
    -------
    chain_order : np.array, shape=(n,), dtype=int
        The permutation of the chains that minimizes the RMSD between the pose and the
        reference.
    transform : AffineTransformation
        The transformation that applied on the pose system gave the minimal RMSD.
    anchor_index : int
        The index of the anchor pose chain.
        This index points to the original order, before `chain_order` is applied.
    """
    # Assign reference and pose entity IDs in a single call
    # in order to assign the same ID to corresponding chains between reference and pose
    entity_ids = _assign_entity_ids(
        reference_chains + pose_chains, min_sequence_identity
    )
    # Split the entity IDs again
    reference_entity_ids = entity_ids[: len(reference_chains)]
    pose_entity_ids = entity_ids[len(reference_chains) :]
    if (
        np.bincount(reference_entity_ids, minlength=len(entity_ids))
        != np.bincount(pose_entity_ids, minlength=len(entity_ids))
    ).any():
        raise ValueError("Reference and pose have different entities")

    anchor_index = _choose_anchor_chain(pose_chains, pose_entity_ids)

    reference_centroids = np.array([struc.centroid(c) for c in reference_chains])
    pose_centroids = np.array([struc.centroid(c) for c in pose_chains])

    best_transform = None
    best_rmsd = np.inf
    best_chain_order = None
    # Test all possible chains that represent the same entity against the anchor chain
    for reference_i, reference_chain in enumerate(reference_chains):
        if reference_entity_ids[reference_i] != pose_entity_ids[anchor_index]:
            continue
        else:
            # Superimpose the entire system
            # based on the anchor and chosen reference chain
            transform = _get_superimposition_transform(
                reference_chain, pose_chains[anchor_index]
            )
            chain_order = _find_matching_centroids(
                reference_centroids,
                pose_centroids,
                reference_entity_ids,
                pose_entity_ids,
                transform,
            )
            superimposed_pose_centroids = transform.apply(pose_centroids)
            rmsd = struc.rmsd(
                reference_centroids, superimposed_pose_centroids[chain_order]
            )
            if rmsd < best_rmsd:
                best_rmsd = rmsd
                best_transform = transform
                best_chain_order = chain_order
    return best_chain_order, best_transform, anchor_index  # type: ignore[return-value]


def _find_matching_centroids(
    reference_centroids: NDArray[np.floating],
    pose_centroids: NDArray[np.floating],
    reference_entity_ids: NDArray[np.int_],
    pose_entity_ids: NDArray[np.int_],
    transform: struc.AffineTransformation,
) -> NDArray[np.int_]:
    """
    Find pairs of chains (each represented by its centroid) between the reference and
    the pose that are closest to each other und the given transformation.

    This functions iteratively chooses pairs with the smallest centroid distance, i.e.
    first the pair with the smallest centroid distance is chosen, then the pair with the
    second smallest centroid distance and so on.

    Parameters
    ----------
    reference_centroids, pose_centroids : np.ndarray, shape=(n,3)
        The centroids of the reference and pose chains.
    reference_entity_ids, pose_entity_ids : np.ndarray, shape=(n,), dtype=int
        The entity IDs of the chains.
        Only centroids of chains with the same entity ID can be matched.
    transform : AffineTransformation
        The transformation that superimposes the pose centroids onto the reference
        centroids.

    Returns
    -------
    pose_order : np.ndarray, shape=(n,)
        The permutation of the pose chains that gives the pairs with the smallest
        distance, i.e. ``pose_order[i] == j`` if the ``i``-th reference chain and
        ``j``-th pose chain are closest to each other.
    """
    pose_centroids = transform.apply(pose_centroids)
    distances = struc.distance(reference_centroids[:, None], pose_centroids[None, :])
    # Different entities must not be matched
    distances[reference_entity_ids[:, None] != pose_entity_ids[None, :]] = np.inf
    pose_order = np.zeros(len(pose_centroids), dtype=int)
    # n chains -> n pairs -> n iterations
    for _ in range(len(pose_centroids)):
        min_distance = np.min(distances)
        min_reference_i, min_pose_i = np.argwhere(distances == min_distance)[0]
        pose_order[min_reference_i] = min_pose_i
        distances[min_reference_i, :] = np.inf
        distances[:, min_pose_i] = np.inf
    return pose_order


def _find_common_residues(
    reference: struc.AtomArray, pose: struc.AtomArray
) -> tuple[NDArray[np.int_], NDArray[np.int_]]:
    """
    Find common residues (and the common atoms) in them in two protein chains.

    Parameters
    ----------
    reference, pose : struc.AtomArray, shape=(n,)
        The reference and pose small molecule, respectively.
        Must have the same atoms (with different coordinates).

    Returns
    -------
    reference_order, pose_order : np.array, shape=(n,), dtype=int
        The atom order that should be applied to the `reference` or `pose`,
        respectively to minimize the RMSD between them.

    Notes
    -----
    Atoms are identified by their element and bond types.
    The element alone is not sufficient, as hydrogen atoms are missing in the
    structures.
    Hence, e.g. a double-bonded oxygen atom would not be distinguishable from a
    hydroxyl group, without the bond types.
    """
    # Shortcut if the structures already match perfectly atom-wise
    if _is_matched(reference, pose, _ANNOTATIONS_FOR_ATOM_MATCHING):
        return np.arange(reference.array_length()), np.arange(reference.array_length())

    reference_sequence = struc.to_sequence(reference)[0][0]
    pose_sequence = struc.to_sequence(pose)[0][0]
    alignment = align.align_optimal(
        reference_sequence,
        pose_sequence,
        _IDENTITY_MATRIX,
        # We get mismatches due to cropping, not due to evolution
        # -> linear gap penalty makes most sense
        gap_penalty=-1,
        max_number=1,
    )[0]
    # Remove gaps -> crop structures to common residues
    alignment.trace = alignment.trace[(alignment.trace != -1).all(axis=1)]

    # Atom masks that are True for atoms in residues that are common in both structures
    reference_mask = _get_mask_from_alignment_trace(reference, alignment.trace[:, 0])
    pose_mask = _get_mask_from_alignment_trace(pose, alignment.trace[:, 1])
    reference_indices = np.arange(reference.array_length())[reference_mask]
    pose_indices = np.arange(pose.array_length())[pose_mask]

    # Within the atoms of aligned residues, select only common atoms
    reference_subset_indices, pose_subset_indices = _find_atom_intersection(
        reference[reference_indices], pose[pose_indices]
    )

    return (
        reference_indices[reference_subset_indices],
        pose_indices[pose_subset_indices],
    )


def _get_mask_from_alignment_trace(
    chain: struc.AtomArray, trace_column: NDArray[np.int_]
) -> NDArray[np.bool_]:
    """
    Get a mask that is True for all atoms, whose residue is contained in the given
    alignment trace column.

    Parameters
    ----------
    chain : AtomArray, shape=(n,)
        The chain to get the mask for.
    trace_column : ndarray, shape=(k,), dtype=int
        The column of the alignment trace for that chain.
        Each index in this trace column points to a residue in the chain.

    Returns
    -------
    mask : ndarray, shape=(n,), dtype=bool
        The mask for the given chain.
    """
    return struc.get_residue_masks(chain, struc.get_residue_starts(chain))[
        trace_column
    ].any(axis=0)


def _find_atom_intersection(
    reference: struc.AtomArray,
    pose: struc.AtomArray,
) -> tuple[NDArray[np.int_], NDArray[np.int_]]:
    """
    Find the intersection of two structures, i.e. the set of equivalent atoms.

    Parameters
    ----------
    reference, pose : AtomArray
        The reference and pose chain, respectively.

    Returns
    -------
    common_reference_indices, common_pose_indices : ndarray, shape=(n,), dtype=int
        The reference and pose indices pointing to the common subset of atoms.

    Notes
    -----
    The order is not necessarily canonical, as the atoms are simply sorted based on the
    given annotations.
    The important requirement is that the order is the same for both structures.
    """
    # Shortcut if the structures already match perfectly atom-wise
    if _is_matched(reference, pose, _ANNOTATIONS_FOR_ATOM_MATCHING):
        return np.arange(reference.array_length()), np.arange(reference.array_length())

    # Use continuous residue IDs to enforce that the later reordering does not mix up
    # atoms from different residues
    reference.res_id = struc.create_continuous_res_ids(reference, False)
    pose.res_id = struc.create_continuous_res_ids(pose, False)
    # Implicitly expect that the annotation array dtypes are the same for both
    structured_dtype = np.dtype(
        [
            (name, pose.get_annotation(name).dtype)
            for name in ["res_id"] + _ANNOTATIONS_FOR_ATOM_MATCHING
        ]
    )
    ref_annotations = _annotations_to_structured(reference, structured_dtype)
    mod_annotations = _annotations_to_structured(pose, structured_dtype)
    ref_indices = np.where(np.isin(ref_annotations, mod_annotations))[0]
    mod_indices = np.where(np.isin(mod_annotations, ref_annotations))[0]
    # Atom ordering might not be same -> sort
    mod_indices = mod_indices[np.argsort(mod_annotations[mod_indices])]
    ref_indices = ref_indices[np.argsort(ref_annotations[ref_indices])]

    return ref_indices, mod_indices


def _annotations_to_structured(
    atoms: struc.AtomArray, structured_dtype: np.dtype
) -> NDArray[Any]:
    """
    Convert atom annotations into a single structured `ndarray`.

    Parameters
    ----------
    atoms : AtomArray
        The annotation arrays are taken from this structure.
    structured_dtype : dtype
        The dtype of the structured array to be created.
        The fields of the dtype determine which annotations are taken from `atoms`.
    """
    if structured_dtype.fields is None:
        raise TypeError("dtype must be structured")
    structured = np.zeros(atoms.array_length(), dtype=structured_dtype)
    for field in structured_dtype.fields:
        structured[field] = atoms.get_annotation(field)
    return structured


def _find_optimal_molecule_permutation(
    reference: struc.AtomArray, pose: struc.AtomArray, superimpose: bool
) -> tuple[NDArray[np.int_], NDArray[np.int_]]:
    """
    Find corresponding atoms in small molecules that minimizes the RMSD between the
    pose and the reference.

    Use graph isomorphism on the bond graph to account for symmetries within
    the small molecule.

    Parameters
    ----------
    reference, pose : struc.AtomArray, shape=(n,)
        The reference and pose small molecule, respectively.
        It is expected that they are already superimposed onto each other.
    superimpose : bool
        Whether to superimpose the reference and the reordered pose onto each other
        before calculating the RMSD.
        This is necessary if no other anchor chain determines the optimal
        superimposition.

    Returns
    -------
    reference_order, pose_order : np.array, shape=(n,), dtype=int
        The atom order that should be applied to the `reference` or `pose`,
        respectively to minimize the RMSD between them.

    Notes
    -----
    Atoms are identified by their element only.
    Currently this allows atoms/groups to be matched even if they are not equivalent
    in some scenarios:

    - Configurational diastereomers (enantiomers and cis/trans isomers)
    - Terminal aldehydes and hydroxyl groups
    """
    reference_mol = _to_mol(reference)
    pose_mol = _to_mol(pose)
    mappings = pose_mol.GetSubstructMatches(
        reference_mol, useChirality=True, uniquify=False
    )
    if len(mappings) == 0:
        raise ValueError(
            "No atom mapping found between pose and reference small molecule"
        )

    best_rmsd = np.inf
    best_pose_atom_order = None
    for pose_atom_order in mappings:
        pose_atom_order = np.array(pose_atom_order)
        if len(pose_atom_order) != reference.array_length():
            raise ValueError("Atom mapping does not cover all atoms")
        if superimpose:
            superimposed, _ = struc.superimpose(reference, pose[pose_atom_order])
        else:
            superimposed = pose[pose_atom_order]
        rmsd = struc.rmsd(reference, superimposed)
        if rmsd < best_rmsd:
            best_rmsd = rmsd
            best_pose_atom_order = pose_atom_order
    return np.arange(reference.array_length()), best_pose_atom_order  # type: ignore[return-value]


def _assign_entity_ids(
    chains: list[struc.AtomArray],
    min_sequence_identity: float,
) -> NDArray[np.int_]:
    """
    Assign a unique entity ID to each distinct chain.

    This means that two chains with the same entity ID have sufficient sequence
    identity or in case of small molecules have the same ``res_name``.

    Parameters
    ----------
    chains : list of struc.AtomArray, length=n
        The chains to assign entity IDs to.
    min_sequence_identity : float
        The minimum sequence identity between two chains to be considered the same
        entity.

    Returns
    -------
    entity_ids : np.ndarray, shape=(n,), dtype=int
        The entity IDs.
    """
    sequences = [
        struc.to_sequence(chain)[0][0] if not is_small_molecule(chain) else None
        for chain in chains
    ]

    current_entity_id = 0
    entity_ids: list[int] = []
    for i, (chain, sequence) in enumerate(zip(chains, sequences)):
        for j in range(i):
            if sequence is None:
                # Match small molecules by residue name
                if chain.res_name[0] == chains[j].res_name[0]:
                    entity_ids.append(entity_ids[j])
                    break
            else:
                # Match protein chains by sequence identity
                alignment = align.align_optimal(
                    sequence,
                    sequences[j],
                    _IDENTITY_MATRIX,
                    # We get mismatches due to experimental artifacts, not evolution
                    # -> linear gap penalty makes most sense
                    gap_penalty=-1,
                    max_number=1,
                )[0]
                if (
                    align.get_sequence_identity(alignment, mode="all")
                    >= min_sequence_identity
                ):
                    entity_ids.append(entity_ids[j])
                    break
        else:
            # No match found to a chain that already has an entity ID -> assign new ID
            entity_ids.append(current_entity_id)
            current_entity_id += 1

    return np.array(entity_ids)


def _choose_anchor_chain(
    chains: list[struc.AtomArray], entity_ids: NDArray[np.int_]
) -> int:
    """
    Choose the anchor chain for the RMSD calculation.

    The most preferable chain is the one with the least multiplicity and the longest
    sequence.

    Parameters
    ----------
    chains : list of struc.AtomArray, length=n
        The chains to choose from.
    entity_ids : ndarray, shape=(n,), dtype=int
        The entity IDs of the chains.
        Used to determine the multiplicity of each chain.

    Returns
    -------
    anchor_chain : int
        The index of the anchor chain.
    """
    protein_chain_indices = np.where(
        [not is_small_molecule(chain) for chain in chains]
    )[0]
    if len(protein_chain_indices) == 0:
        # No protein chains -> Simply use the first small molecule as anchor
        return 0

    protein_entity_ids = entity_ids[protein_chain_indices]
    multiplicities_of_entity_ids = np.bincount(protein_entity_ids)
    multiplicities = multiplicities_of_entity_ids[protein_entity_ids]
    least_multiplicity_indices = np.where(multiplicities == np.min(multiplicities))[0]
    # Use the sequence length as tiebreaker
    sequence_lengths = np.array([len(chains[i]) for i in protein_chain_indices])
    # Only consider the lengths of the preselected chains
    largest_length = np.max(sequence_lengths[least_multiplicity_indices])
    largest_length_indices = np.where(sequence_lengths == largest_length)[0]
    best_anchors = np.intersect1d(least_multiplicity_indices, largest_length_indices)
    return protein_chain_indices[best_anchors[0]]


def _get_superimposition_transform(
    reference_chain: struc.AtomArray, pose_chain: struc.AtomArray
) -> struc.AffineTransformation:
    """
    Get the transformation (translation and rotation) that superimposes the pose chain
    onto the reference chain.

    Parameters
    ----------
    reference_chain, pose_chain : AtomArray
        The chains to superimpose.

    Returns
    -------
    transform : AffineTransformation
        The transformation that superimposes the pose chain onto the reference chain.
    """
    if is_small_molecule(reference_chain):
        if reference_chain.array_length() == pose_chain.array_length():
            _, transform = struc.superimpose(reference_chain, pose_chain)
        else:
            # The small molecules have different lengths -> difficult superimposition
            # -> simply get an identity transformation
            # This case can only happen in small molecule-only systems anyway
            transform = struc.AffineTransformation(
                center_translation=np.zeros(3),
                rotation=np.eye(3),
                target_translation=np.zeros(3),
            )
    else:
        _, transform, _, _ = struc.superimpose_homologs(
            reference_chain,
            pose_chain,
            _IDENTITY_MATRIX,
            gap_penalty=-1,
            min_anchors=1,
            # No outlier removal
            max_iterations=1,
        )
    return transform


def _is_matched(
    reference: struc.AtomArray,
    pose: struc.AtomArray,
    annotation_names: list[str],
) -> bool:
    """
    Check if the given annotations are the same in both structures.

    Parameters
    ----------
    pose, reference : AtomArray
        The pose and reference structure to be compared, respectively.
    annotation_names : list of str
        The names of the annotations to be compared.

    Returns
    -------
    matched : bool
        True, if the annotations are the same in both structures.
    """
    if reference.array_length() != pose.array_length():
        return False
    for annot_name in annotation_names:
        if not (
            reference.get_annotation(annot_name) == pose.get_annotation(annot_name)
        ).all():
            return False
    return True


def _to_mol(molecule: struc.AtomArray) -> Mol:
    """
    Create a RDKit :class:`Mol` from the given structure and prepare it for usage in
    atom matching.

    Parameters
    ----------
    molecule : struc.AtomArray
        The molecule to convert.

    Returns
    -------
    mol : Mol
        The RDKit molecule.
    """
    mol = rdkit_interface.to_mol(molecule)
    # Make RDKit distinguish stereoisomers when matching atoms
    AssignStereochemistryFrom3D(mol)
    # Make conjugated terminal groups symmetric (e.g. carboxyl oxygen atoms)
    SanitizeMol(mol, SanitizeFlags.SANITIZE_SETCONJUGATION)
    for bond in mol.GetBonds():
        if bond.GetIsConjugated() and bond.GetBondType() in [
            BondType.SINGLE,
            BondType.DOUBLE,
        ]:
            bond.SetBondType(BondType.ONEANDAHALF)
    return mol



"""
Calculate DockQ for a single pair of receptor and ligand.
"""

__all__ = [
    "get_contact_residues",
    "dockq",
    "pocket_aligned_lrmsd",
    "lrmsd",
    "irmsd",
    "fnat",
    "DockQ",
    "NoContactError",
]

from dataclasses import dataclass, field
from typing import overload
import biotite.structure as struc
import numpy as np
from numpy.typing import NDArray
from peppr.common import is_small_molecule

BACKBONE_ATOMS = (
    "CA",
    "C",
    "N",
    "O",
    "P",
    "OP1",
    "OP2",
    "O2'",
    "O3'",
    "O4'",
    "O5'",
    "C1'",
    "C2'",
    "C3'",
    "C4'",
    "C5'",
)


class NoContactError(Exception):
    pass


@dataclass(frozen=True)
class DockQ:
    """
    Result of a *DockQ* calculation.

    If multiple poses were used to calculate *DockQ*, the attributes are arrays.

    Attributes
    ----------
    fnat : float or ndarray, dtype=float
        The fraction of reference contacts found in the pose relative to the total
        number of reference contacts.
    fnonnat : float or ndarray, dtype=float
        The fraction of non-reference contacts found in the pose relative to the total
        number of pose contacts.
    irmsd : float or ndarray, dtype=float
        The interface RMSD.
    lrmsd : float or ndarray, dtype=float
        The ligand RMSD.
    score : float or ndarray, dtype=float
        The DockQ score.
    n_poses : int or None
        The number of poses for which the *DockQ* was calculated.
        `None`, if the *DockQ* was calculated for an `AtomArray`.
    pose_receptor_index, pose_ligand_index, reference_receptor_index, reference_ligand_index : int or None
        The indices of the pose and reference chain that were included for *DockQ*
        computation.
        Only set, if called from `global_dockq()`.
    """

    fnat: float | NDArray[np.floating]
    fnonnat: float | NDArray[np.floating]
    irmsd: float | NDArray[np.floating]
    lrmsd: float | NDArray[np.floating]
    score: float | NDArray[np.floating] = field(init=False)
    n_poses: int | None = field(init=False)
    pose_receptor_index: int | None = None
    pose_ligand_index: int | None = None
    reference_receptor_index: int | None = None
    reference_ligand_index: int | None = None

    def __post_init__(self) -> None:
        score = np.nanmean(
            [self.fnat, _scale(self.irmsd, 1.5), _scale(self.lrmsd, 8.5)], axis=0
        )
        n_poses = None if np.isscalar(score) else len(score)
        super().__setattr__("score", score)
        super().__setattr__("n_poses", n_poses)

    def for_pose(self, pose_index: int) -> "DockQ":
        """
        Get the DockQ results for a specific pose index.

        Parameters
        ----------
        pose_index : int
            The index of the pose for which the DockQ results should be retrieved.

        Returns
        -------
        DockQ
            The DockQ results for the specified pose index.

        Raises
        ------
        IndexError
            If the `GlobalDockQ` object was computed for a single pose,
            i.e. `n_poses` is `None`.
        """
        if self.n_poses is None:
            raise IndexError("DockQ was computed for a single pose")
        return DockQ(
            self.fnat[pose_index].item(),  # type: ignore[index]
            self.fnonnat[pose_index].item(),  # type: ignore[index]
            self.irmsd[pose_index].item(),  # type: ignore[index]
            self.lrmsd[pose_index].item(),  # type: ignore[index]
            pose_receptor_index=self.pose_receptor_index,
            pose_ligand_index=self.pose_ligand_index,
            reference_receptor_index=self.reference_receptor_index,
            reference_ligand_index=self.reference_ligand_index,
        )


def dockq(
    reference_receptor: struc.AtomArray,
    reference_ligand: struc.AtomArray,
    pose_receptor: struc.AtomArray | struc.AtomArrayStack,
    pose_ligand: struc.AtomArray | struc.AtomArrayStack,
    as_peptide: bool = False,
) -> DockQ:
    """
    Compute *DockQ* for a single pair of receptor and ligand in both, the pose and
    reference structure.

    Parameters
    ----------
    reference_receptor, reference_ligand : AtomArray
        The reference receptor and ligand.
    pose_receptor, pose_ligand : AtomArray or AtomArrayStack
        The pose receptor and ligand.
        Multiple poses can be provided.
    as_peptide : bool
        If set to true, the chains are treated as CAPRI peptides.

    Returns
    -------
    DockQ
        The DockQ result.
        If multiple poses are provided, the `DockQ` attributes are arrays.

    Notes
    -----
    If the ligand is a small molecule, an associated `BondList` is required in
    `pose_ligand` and `reference_ligand` for mapping the atoms between them.

    Examples
    --------

    Single chains as expected as input.

    >>> pose_receptor = pose_complex[pose_complex.chain_id == "C"]
    >>> pose_ligand = pose_complex[pose_complex.chain_id == "B"]
    >>> reference_receptor = reference_complex[reference_complex.chain_id == "C"]
    >>> reference_ligand = reference_complex[reference_complex.chain_id == "B"]
    >>> dockq_result = dockq(pose_receptor, pose_ligand, reference_receptor, reference_ligand)
    >>> print(f"{dockq_result.fnat:.2f}")
    0.50
    >>> print(f"{dockq_result.irmsd:.2f}")
    2.10
    >>> print(f"{dockq_result.lrmsd:.2f}")
    8.13
    >>> print(f"{dockq_result.score:.2f}")
    0.45
    """
    undefined = (
        np.nan
        if isinstance(pose_receptor, struc.AtomArray)
        else np.full(pose_receptor.stack_depth(), np.nan)
    )

    if as_peptide:
        if any(
            [
                is_small_molecule(chain)
                for chain in (
                    pose_receptor,
                    pose_ligand,
                    reference_receptor,
                    reference_ligand,
                )
            ]
        ):
            raise ValueError("'as_peptide' is true, but the chains are not peptides")

    if is_small_molecule(pose_ligand):
        # For small molecules DockQ is only based on the pocket-aligned ligand RMSD
        lrmsd_ = pocket_aligned_lrmsd(
            reference_receptor, reference_ligand, pose_receptor, pose_ligand
        )
        return DockQ(undefined, undefined, undefined, lrmsd_)

    else:
        fnat_, fnonnat_ = fnat(
            reference_receptor,
            reference_ligand,
            pose_receptor,
            pose_ligand,
            as_peptide,
        )
        if np.isnan(fnat_).any():
            # No contact between the chains -> DockQ is not defined
            return DockQ(undefined, undefined, undefined, undefined)

        irmsd_ = irmsd(
            reference_receptor,
            reference_ligand,
            pose_receptor,
            pose_ligand,
            as_peptide,
        )

        lrmsd_ = lrmsd(reference_receptor, reference_ligand, pose_receptor, pose_ligand)

        return DockQ(fnat_, fnonnat_, irmsd_, lrmsd_)


def pocket_aligned_lrmsd(
    reference_receptor: struc.AtomArray,
    reference_ligand: struc.AtomArray,
    pose_receptor: struc.AtomArray | struc.AtomArrayStack,
    pose_ligand: struc.AtomArray | struc.AtomArrayStack,
) -> float | NDArray[np.floating]:
    """
    Compute the pocket-aligned RMSD part of the DockQ score for small molecules.

    Parameters
    ----------
    reference_receptor, reference_ligand : AtomArray
        The reference receptor and ligand.
    pose_receptor, pose_ligand : AtomArray
        The pose receptor and ligand.

    Returns
    -------
    float or ndarray, dtype=float
        The pocket-aligned RMSD.
    """
    reference_contacts = get_contact_residues(
        reference_receptor,
        reference_ligand,
        cutoff=10.0,
    )
    if len(reference_contacts) == 0:
        # if there're no contacts between the two chains, no lrmsd is calculated
        return (
            np.full(shape=len(pose_ligand), fill_value=np.nan)
            if isinstance(pose_ligand, struc.AtomArrayStack)
            else np.nan
        )
    # Create mask which is True for all backbone atoms in contact receptor residues
    interface_mask = struc.get_residue_masks(
        reference_receptor, reference_contacts[:, 0]
    ).any(axis=0) & np.isin(reference_receptor.atom_name, BACKBONE_ATOMS)
    # Use interface backbone coordinates for pocket-aligned superimposition
    _, transform = struc.superimpose(
        reference_receptor.coord[interface_mask],
        pose_receptor.coord[..., interface_mask, :],
    )
    # Use the superimposed coordinates for RMSD calculation between ligand atoms
    lrmsd = struc.rmsd(reference_ligand.coord, transform.apply(pose_ligand.coord))
    return lrmsd.item() if np.isscalar(lrmsd) else lrmsd  # type: ignore[union-attr]


def lrmsd(
    reference_receptor: struc.AtomArray,
    reference_ligand: struc.AtomArray,
    pose_receptor: struc.AtomArray | struc.AtomArrayStack,
    pose_ligand: struc.AtomArray | struc.AtomArrayStack,
) -> float | NDArray[np.floating]:
    """
    Compute the ligand RMSD part of the DockQ score.

    Parameters
    ----------
    reference_receptor, reference_ligand : AtomArray
        The reference receptor and ligand.
    pose_receptor, pose_ligand : AtomArray
        The pose receptor and ligand.

    Returns
    -------
    float or ndarray, dtype=float
        The ligand RMSD.
    """
    receptor_relevant_mask = np.isin(pose_receptor.atom_name, BACKBONE_ATOMS)
    if is_small_molecule(pose_ligand):
        # For small molecules include all heavy atoms
        ligand_relevant_mask = np.full(pose_ligand.array_length(), True)
    else:
        ligand_relevant_mask = np.isin(pose_ligand.atom_name, BACKBONE_ATOMS)

    pose_receptor_coord = pose_receptor.coord[..., receptor_relevant_mask, :]
    pose_ligand_coord = pose_ligand.coord[..., ligand_relevant_mask, :]
    reference_receptor_coord = reference_receptor.coord[receptor_relevant_mask]
    reference_ligand_coord = reference_ligand.coord[ligand_relevant_mask]
    _, transform = struc.superimpose(
        reference_receptor_coord,
        pose_receptor_coord,
    )
    superimposed_ligand_coord = transform.apply(pose_ligand_coord)
    lrmsd = struc.rmsd(reference_ligand_coord, superimposed_ligand_coord)
    return lrmsd.item() if np.isscalar(lrmsd) else lrmsd  # type: ignore[union-attr]


def irmsd(
    reference_receptor: struc.AtomArray,
    reference_ligand: struc.AtomArray,
    pose_receptor: struc.AtomArray | struc.AtomArrayStack,
    pose_ligand: struc.AtomArray | struc.AtomArrayStack,
    as_peptide: bool = False,
) -> float | NDArray[np.floating]:
    """
    Compute the interface RMSD part of the DockQ score.

    Parameters
    ----------
    reference_receptor, reference_ligand : AtomArray
        The reference receptor and ligand.
    pose_receptor, pose_ligand : AtomArray or AtomArrayStack
        The pose receptor and ligand.
    as_peptide : bool
        If set to true, the chains are treated as CAPRI peptides.

    Returns
    -------
    float or ndarray, dtype=float
        The interface RMSD.
    """
    if as_peptide:
        cutoff = 8.0
        receptor_mask = _mask_either_or(reference_receptor, "CB", "CA")
        ligand_mask = _mask_either_or(reference_ligand, "CB", "CA")
    else:
        cutoff = 10.0
        receptor_mask = None
        ligand_mask = None
    reference_contacts = get_contact_residues(
        reference_receptor,
        reference_ligand,
        cutoff,
        receptor_mask,
        ligand_mask,
    )

    if len(reference_contacts) == 0:
        # if there're no contacts between the two chains,
        # no irmsd is calculated
        return (
            np.full(shape=len(pose_ligand), fill_value=np.nan)
            if isinstance(pose_ligand, struc.AtomArrayStack)
            else np.nan
        )

    # Create mask which is True for all backbone atoms in contact residues
    receptor_backbone_interface_mask = struc.get_residue_masks(
        reference_receptor, reference_contacts[:, 0]
    ).any(axis=0) & np.isin(reference_receptor.atom_name, BACKBONE_ATOMS)
    ligand_backbone_interface_mask = struc.get_residue_masks(
        reference_ligand, reference_contacts[:, 1]
    ).any(axis=0) & np.isin(reference_ligand.atom_name, BACKBONE_ATOMS)

    # Get the coordinates of interface backbone atoms
    pose_interface_coord = np.concatenate(
        [
            pose_receptor.coord[..., receptor_backbone_interface_mask, :],
            pose_ligand.coord[..., ligand_backbone_interface_mask, :],
        ],
        axis=-2,
    )
    reference_interface_coord = np.concatenate(
        [
            reference_receptor.coord[receptor_backbone_interface_mask],
            reference_ligand.coord[ligand_backbone_interface_mask],
        ],
        axis=-2,
    )
    # Use these coordinates for superimposition and RMSD calculation
    superimposed_interface_coord, _ = struc.superimpose(
        reference_interface_coord, pose_interface_coord
    )
    rmsd = struc.rmsd(reference_interface_coord, superimposed_interface_coord)
    return rmsd.item() if np.isscalar(rmsd) else rmsd  # type: ignore[union-attr]


def fnat(
    reference_receptor: struc.AtomArray,
    reference_ligand: struc.AtomArray,
    pose_receptor: struc.AtomArray | struc.AtomArrayStack,
    pose_ligand: struc.AtomArray | struc.AtomArrayStack,
    as_peptide: bool = False,
) -> tuple[float | NDArray[np.floating], float | NDArray[np.floating]]:
    """
    Compute the *fnat* and *fnonnat* part of the DockQ score.

    Parameters
    ----------
    reference_receptor, reference_ligand : AtomArray
        The reference receptor and ligand.
    pose_receptor, pose_ligand : AtomArray or AtomArrayStack
        The pose receptor and ligand.
    as_peptide : bool
        If set to true, the chains are treated as CAPRI peptides.

    Returns
    -------
    fnat : float or ndarray, dtype=float
        The percentage of reference contacts that are also found in the pose.
    fnonnat : float or ndarray, dtype=float
        The percentage of pose contacts that are not found in the reference structure.
    """
    cutoff = 4.0 if as_peptide else 5.0

    reference_contacts = _as_set(
        get_contact_residues(reference_receptor, reference_ligand, cutoff)
    )
    if len(reference_contacts) == 0:
        # if there're no contacts between the two chains, fnat and fnonnat are not defined
        nan_values = (
            np.full(shape=len(pose_ligand), fill_value=np.nan)
            if isinstance(pose_ligand, struc.AtomArrayStack)
            else np.nan
        )
        return nan_values, nan_values

    if isinstance(pose_receptor, struc.AtomArray):
        return _calc_fnat_single_model(
            pose_receptor,
            pose_ligand,
            reference_contacts,
            cutoff,
        )
    else:
        fnat = []
        fnonnat = []
        # Multiple poses in an AtomArrayStack -> calculate fnat for each pose
        for receptor, ligand in zip(pose_receptor, pose_ligand):
            fnat_single, fnonnat_single = _calc_fnat_single_model(
                receptor,
                ligand,
                reference_contacts,
                cutoff,
            )
            fnat.append(fnat_single)
            fnonnat.append(fnonnat_single)
        return np.array(fnat, dtype=float), np.array(fnonnat, dtype=float)


def get_contact_residues(
    receptor: struc.AtomArray,
    ligand: struc.AtomArray,
    cutoff: float,
    receptor_mask: NDArray[np.int_] | None = None,
    ligand_mask: NDArray[np.int_] | None = None,
) -> NDArray[np.int_]:
    """
    Get a set of tuples containing the residue IDs for each contact between
    receptor and ligand.

    Parameters
    ----------
    receptor, ligand : AtomArray, shape=(p,)
        The receptor.
    ligand : AtomArray, shape=(q,)
        The ligand.
    cutoff : float
        The distance cutoff for contact.
    receptor_mask : ndarray, shape=(p,), dtype=bool, optional
        A mask that is `True` for atoms in `receptor` that should be considered.
        If `None`, all atoms are considered.
    ligand_mask : ndarray, shape=(q,), dtype=bool, optional
        A mask that is `True` for atoms in `ligand` that should be considered.
        If `None`, all atoms are considered.

    Returns
    -------
    ndarray, shape=(n,2), dtype=int
        Each row represents a contact between receptor and ligand.
        The first column contains the starting atom index of the receptor residue,
        the second column contains the starting atom index of the ligand residue.
    """
    # Put the receptor instead of the ligand into the cell list
    # as the receptor is usually larger
    # This increases the performance in CellList.get_atoms() and _to_sparse_indices()
    cell_list = struc.CellList(receptor, cutoff, selection=receptor_mask)
    if ligand_mask is None:
        all_contacts = cell_list.get_atoms(ligand.coord, cutoff)
    else:
        filtered_contacts = cell_list.get_atoms(ligand.coord[ligand_mask], cutoff)
        all_contacts = np.full(
            (len(ligand), filtered_contacts.shape[-1]),
            -1,
            dtype=filtered_contacts.dtype,
        )
        all_contacts[ligand_mask] = filtered_contacts
    atom_indices = _to_sparse_indices(all_contacts)

    residue_starts = np.stack(
        [
            struc.get_residue_starts_for(receptor, atom_indices[:, 0]),
            struc.get_residue_starts_for(ligand, atom_indices[:, 1]),
        ],
        axis=1,
    )

    # Some contacts might exist between different atoms in the same residues
    return np.unique(residue_starts, axis=0)


def _calc_fnat_single_model(
    receptor: struc.AtomArray,
    ligand: struc.AtomArray,
    reference_contacts: set[tuple[int, int]],
    cutoff: float,
) -> tuple[float | NDArray[np.floating], float | NDArray[np.floating]]:
    """
    Compute the *fnat* and *fnonnat* for a single model.

    Parameters
    ----------
    receptor, ligand : AtomArray
        The pose receptor and ligand.
    reference_contacts : ndarray, shape=(n,2), dtype=int
        The reference contacts.
    cutoff : float
        The distance cutoff for contact.

    Returns
    -------
    fnat : float
        The percentage of reference contacts that are also found in the pose.
    fnonnat : float
        The percentage of pose contacts that are not found in the reference structure.
    """
    pose_contacts = _as_set(get_contact_residues(receptor, ligand, cutoff))
    n_pose = len(pose_contacts)
    n_reference = len(reference_contacts)
    n_true_positive = len(pose_contacts & reference_contacts)
    n_false_positive = len(pose_contacts - reference_contacts)

    if n_reference == 0:
        # Deviation from original DockQ implementation, which returns 0 in this case
        # However, this is misleading as the score is simply not properly defined
        # in this case, as the structure is not a complex
        raise NoContactError("The reference chains do not have any contacts")
    fnat = n_true_positive / n_reference
    fnonnat = n_false_positive / n_pose if n_pose != 0 else 0
    return fnat, fnonnat


def _mask_either_or(
    atoms: struc.AtomArray, atom_name: str, alt_atom_name: str
) -> NDArray[np.bool_]:
    """
    Create a mask that is `True` for all `atom_name` atoms and as fallback
    `alt_atom_name` atoms for all residues that miss `atom_name`.

    Parameters
    ----------
    atoms : AtomArray, shape=(n,)
        The atoms to create the mask for.
    atom_name : str
        The atom name to base the mask on.
    alt_atom_name : str
        The fallback atom name.

    Returns
    -------
    mask : ndarray, shape=(n,), dtype=bool
        The created mask.
    """
    atom_names = atoms.atom_name
    mask = np.zeros(atoms.array_length(), dtype=bool)
    residue_starts = struc.get_residue_starts(atoms, add_exclusive_stop=True)
    for i in range(len(residue_starts) - 1):
        res_start = residue_starts[i]
        res_stop = residue_starts[i + 1]
        atom_index = np.where(atom_names[res_start:res_stop] == atom_name)[0]
        if len(atom_index) != 0:
            mask[res_start + atom_index] = True
        else:
            # No `atom_name` in residue -> fall back to `alt_atom_name`
            atom_index = np.where(atom_names[res_start:res_stop] == alt_atom_name)[0]
            if len(atom_index) != 0:
                mask[res_start + atom_index] = True
    return mask


def _as_set(array: NDArray[np.int_]) -> set[tuple[int, int]]:
    """
    Convert an array of tuples into a set of tuples.
    """
    return set([tuple(c) for c in array])


def _to_sparse_indices(all_contacts: NDArray[np.int_]) -> NDArray[np.int_]:
    """
    Create tuples of indices that would mark the non-zero elements in a dense
    contact matrix.
    """
    # Find rows where a ligand atom has at least one contact
    non_empty_indices = np.where(np.any(all_contacts != -1, axis=1))[0]
    # Take those rows and flatten them
    receptor_indices = all_contacts[non_empty_indices].flatten()
    # For each row the corresponding ligand atom is the same
    # Hence in the flattened form the ligand atom index is simply repeated
    ligand_indices = np.repeat(non_empty_indices, all_contacts.shape[1])
    combined_indices = np.stack([receptor_indices, ligand_indices], axis=1)
    # Remove the padding values
    return combined_indices[receptor_indices != -1]


@overload
def _scale(rmsd: float, scaling_factor: float) -> float: ...
@overload
def _scale(
    rmsd: NDArray[np.floating], scaling_factor: float
) -> NDArray[np.floating]: ...
def _scale(
    rmsd: float | NDArray[np.floating], scaling_factor: float
) -> float | NDArray[np.floating]:
    """
    Apply the DockQ scaling formula.
    """
    return 1 / (1 + (rmsd / scaling_factor) ** 2)



__all__ = ["Evaluator", "MatchWarning", "EvaluationWarning"]

import copy
import warnings
from collections.abc import Iterable, Mapping
from typing import Iterator, Sequence
import biotite.structure as struc
import numpy as np
import pandas as pd
from peppr.common import standardize
from peppr.match import find_matching_atoms
from peppr.metric import Metric
from peppr.selector import Selector


class MatchWarning(UserWarning):
    """
    This warning is raised, if a the :class:`Evaluator` fails to match atoms between
    the reference and pose structures.
    """

    pass


class EvaluationWarning(UserWarning):
    """
    This warning is raised, if a :class:`Metric` fails to evaluate a pose.
    """

    pass


class Evaluator(Mapping):
    """
    This class represents the core of :mod:`peppr`.
    Systems are fed via :meth:`feed()` into the :class:`Evaluator`.
    Finally, the evaluation is reported via :meth:`tabulate_metrics()`, which gives a
    scalar metric value for each fed system, or via :meth:`summarize_metrics()`,
    which aggregates the metrics over all systems.

    Parameters
    ----------
    metrics : Iterable of Metric
        The metrics to evaluate the poses against.
        These will make up the columns of the resulting dataframe from
        :meth:`tabulate_metrics()`.
    tolerate_exceptions : bool, optional
        If set to true, exceptions during :class:`Metric.evaluate()` are not propagated.
        Instead a warning is raised and the result is set to ``None``.
    min_sequence_identity : float
        The minimum sequence identity for two chains to be considered the same entity.

    Attributes
    ----------
    metrics : tuple of Metric
        The metrics to evaluate the poses against.
    system_ids : tuple of str
        The IDs of the systems that were fed into the evaluator.
    """

    def __init__(
        self,
        metrics: Iterable[Metric],
        tolerate_exceptions: bool = False,
        min_sequence_identity: float = 0.95,
    ):
        self._metrics = tuple(metrics)
        self._results: list[list[np.ndarray]] = [[] for _ in range(len(metrics))]
        self._ids: list[str] = []
        self._tolerate_exceptions = tolerate_exceptions
        self._min_sequence_identity = min_sequence_identity

    @property
    def metrics(self) -> tuple[Metric, ...]:
        # Use tuple to forbid adding/removing metrics after initialization
        return self._metrics

    @property
    def system_ids(self) -> tuple[str, ...]:
        return tuple(self._ids)

    def feed(
        self,
        system_id: str,
        reference: struc.AtomArray,
        poses: Sequence[struc.AtomArray] | struc.AtomArrayStack | struc.AtomArray,
    ) -> None:
        """
        Evaluate the poses of a system against the reference structure for all metrics.

        Parameters
        ----------
        system_id : str
            The ID of the system that was evaluated.
        reference : AtomArray
            The reference structure of the system.
            Each separate instance/molecule must have a distinct `chain_id`.
        poses : AtomArrayStack or list of AtomArray or AtomArray
            The pose(s) to evaluate.
            It is expected that the poses are sorted from highest to lowest confidence,
            (relevant for :class:`Selector` instances).

        Notes
        -----
        `reference` and `poses` must fulfill the following requirements:

        - The system must have an associated `biotite.structure.BondList`,
          i.e. the ``bonds`` attribute must not be ``None``.
        - Each molecule in the system must have a distinct ``chain_id``.
        - Chains where the ``hetero`` annotation is ``True`` is always interpreted as a
          small molecule.
          Conversely, chains where the ``hetero`` annotation is ``False`` is always
          interpreted as protein or nucleic acid chain.

        The optimal chain mapping and atom mapping in symmetric small molecules
        is handled automatically.
        """
        reference = standardize(reference)
        if isinstance(poses, struc.AtomArray):
            poses = [standardize(poses)]
        elif isinstance(poses, struc.AtomArrayStack):
            poses = list(standardize(poses))
        else:
            poses = [standardize(pose) for pose in poses]
        if len(poses) == 0:
            raise ValueError("No poses provided")

        result_for_system = np.full((len(self._metrics), len(poses)), np.nan)
        for j, pose in enumerate(poses):
            try:
                reference_order, pose_order = find_matching_atoms(
                    reference, pose, min_sequence_identity=self._min_sequence_identity
                )
                matched_reference = reference[reference_order]
                matched_pose = pose[pose_order]
            except Exception as e:
                if self._tolerate_exceptions:
                    warnings.warn(
                        f"Failed to match reference and pose in system '{system_id}': {e}",
                        MatchWarning,
                    )
                    continue
                else:
                    raise
            # Sanity check if something went wrong in the matching
            if not np.array_equal(matched_reference.element, matched_pose.element):
                if self._tolerate_exceptions:
                    warnings.warn(
                        f"'{system_id}' poses could not be matched to reference",
                        EvaluationWarning,
                    )
                else:
                    raise ValueError(
                        f"'{system_id}' poses could not be matched to reference"
                    )

            for i, metric in enumerate(self._metrics):
                try:
                    result_for_system[i, j] = metric.evaluate(
                        matched_reference, matched_pose
                    )
                except Exception as e:
                    if self._tolerate_exceptions:
                        warnings.warn(
                            f"Failed to evaluate {metric.name} on '{system_id}': {e}",
                            EvaluationWarning,
                        )
                    else:
                        raise
        for i, result in enumerate(result_for_system):
            self._results[i].append(result)
        self._ids.append(system_id)

    def get_results(self) -> list[list[np.ndarray]]:
        """
        Return the raw results of the evaluation.

        This includes each metric evaluated on each pose of each system.

        Returns
        -------
        list of list of np.ndarray
            The raw results of the evaluation.
            The outer list iterates over the metrics, the inner list iterates over
            the systems and the array represents the values for each pose.
        """
        return copy.deepcopy(self._results)

    def tabulate_metrics(
        self, selectors: Iterable[Selector] | None = None
    ) -> pd.DataFrame:
        """
        Create a table listing the value for each metric and system.

        Parameters
        ----------
        selectors : list of Selector, optional
            The selectors to use for selecting the best pose of a multi-pose
            prediction.
            This parameter is not necessary if only single-pose predictions were fed
            into the :class:`Evaluator`.

        Returns
        -------
        pandas.DataFrame
            A table listing the value for each metric and system.
            The index is the system ID.

        Examples
        --------

        >>> print(evaluator.tabulate_metrics())
                                           RMSD      lDDT  TM-score
        8ji2__1__1.B__1.J_1.K          2.987937  0.674205  0.883589
        7t4w__1__1.A__1.C             16.762669  0.693087  0.380107
        8jp0__1__1.A__1.B             26.281593  0.510061  0.316204
        7yn2__1__1.A_1.B__1.C          6.657655  0.567117  0.725322
        8oxu__2__1.C__1.E             14.977116  0.339707  0.296535
        7ydq__1__1.A__1.B             26.111820  0.383841  0.360584
        7wuy__1__1.B__1.HA_1.IA_1.OA  16.494774  0.665949  0.666633
        7xh4__1__1.A__1.B_1.C          1.787062  0.748987  0.915388
        7v34__1__1.A__1.C_1.D_1.G      4.472874  0.567491  0.822537
        8jmr__1__1.A_1.B__1.C_1.D      5.058327  0.730324  0.868684
        >>> print(evaluator.tabulate_metrics(OracleSelector()))
                                      CA-RMSD (Oracle)  lDDT (Oracle)  TM-score (Oracle)
        8ji2__1__1.B__1.J_1.K                 2.987937       0.674205           0.883589
        7t4w__1__1.A__1.C                    16.762669       0.693087           0.380107
        8jp0__1__1.A__1.B                    26.281593       0.510061           0.316204
        7yn2__1__1.A_1.B__1.C                 6.657655       0.567117           0.725322
        8oxu__2__1.C__1.E                    14.977116       0.339707           0.296535
        7ydq__1__1.A__1.B                    26.111820       0.383841           0.360584
        7wuy__1__1.B__1.HA_1.IA_1.OA         16.494774       0.665949           0.666633
        7xh4__1__1.A__1.B_1.C                 1.787062       0.748987           0.915388
        7v34__1__1.A__1.C_1.D_1.G             4.472874       0.567491           0.822537
        8jmr__1__1.A_1.B__1.C_1.D             5.058327       0.730324           0.868684
        """
        columns = self._tabulate_metrics(selectors)
        # Convert (metric, selector)-tuples to strings
        columns = {
            (
                f"{metric.name} ({selector.name})"
                if selector is not None
                else metric.name
            ): values
            for (metric, selector), values in columns.items()
        }
        return pd.DataFrame(columns, index=self._ids)

    def summarize_metrics(
        self, selectors: Iterable[Selector] | None = None
    ) -> dict[str, float]:
        """
        Condense the system-wise evaluation to scalar values for each metric.

        For each metric,

        - the mean value
        - the median value
        - and the percentage of systems within each threshold

        is computed.

        Parameters
        ----------
        selectors : list of Selector, optional
            The selectors to use for selecting the best pose of a multi-pose
            prediction.
            This parameter is not necessary if only single-pose predictions were fed
            into the :class:`Evaluator`.

        Returns
        -------
        dict (str -> float)
            A dictionary mapping the summarized metric name to the scalar value.
            The summarized metric name contains

            - the metric name (e.g. ``DockQ``)
            - the selector name, if a selector was used (e.g. ``Oracle``)
            - the threshold (if a threshold was used) (e.g. ``% acceptable``)

        Examples
        --------

        >>> import pprint
        >>> pprint.pprint(evaluator.summarize_metrics())
        {'CA-RMSD <5.0': 0.3,
         'CA-RMSD >5.0': 0.7,
         'CA-RMSD mean': 12.159182685504375,
         'TM-score mean': 0.6235582438144873,
         'lDDT mean': 0.5880769924413414}
        >>> pprint.pprint(evaluator.summarize_metrics([MeanSelector(), OracleSelector()]))
        {'CA-RMSD <5.0 (Oracle)': 0.3,
         'CA-RMSD <5.0 (mean)': 0.3,
         'CA-RMSD >5.0 (Oracle)': 0.7,
         'CA-RMSD >5.0 (mean)': 0.7,
         'CA-RMSD mean (Oracle)': 12.159182685504375,
         'CA-RMSD mean (mean)': 12.159182685504375,
         'TM-score mean (Oracle)': 0.6235582438144873,
         'TM-score mean (mean)': 0.6235582438144873,
         'lDDT mean (Oracle)': 0.5880769924413414,
         'lDDT mean (mean)': 0.5880769924413414}
        """
        columns = self._tabulate_metrics(selectors)
        output_columns = {}
        for (metric, selector), values in columns.items():
            if metric.thresholds:
                edges = list(metric.thresholds.values()) + [np.inf]
                counts_per_bin, _ = np.histogram(values, bins=edges)
                # NaN values do not bias the percentages,
                # as they are not included in any bin
                percentages_per_bin = counts_per_bin / np.sum(counts_per_bin)
                for threshold_name, percentage in zip(
                    metric.thresholds.keys(), percentages_per_bin
                ):
                    column_name = f"{metric.name} {threshold_name}"
                    if selector is not None:
                        column_name += f" ({selector.name})"
                    output_columns[column_name] = percentage.item()
            # Always add the mean and median value as well
            for name, function in [("mean", np.nanmean), ("median", np.nanmedian)]:
                column_name = f"{metric.name} {name}"
                if selector is not None:
                    column_name += f" ({selector.name})"
                output_columns[column_name] = function(values).item()  # type: ignore[operator]
        return output_columns

    def _tabulate_metrics(
        self, selectors: Iterable[Selector] | None = None
    ) -> dict[tuple[Metric, Selector | None], np.ndarray]:
        columns: dict[tuple[Metric, Selector | None], np.ndarray] = {}
        for i, metric in enumerate(self._metrics):
            values = self._results[i]
            if not selectors:
                condensed_values = []
                for array in values:
                    if array is None:
                        condensed_values.append(np.nan)
                    elif len(array) > 1:
                        raise ValueError(
                            "At least one selector is required for multi-pose predictions"
                        )
                    else:
                        condensed_values.append(array[0])
                columns[metric, None] = np.array(condensed_values)
            else:
                for selector in selectors:
                    condensed_values = np.array(
                        [
                            selector.select(val, metric.smaller_is_better())
                            if val is not None
                            else np.nan
                            for val in values
                        ]
                    )  # type: ignore[assignment]
                    columns[metric, selector] = condensed_values  # type: ignore[assignment]
        return columns

    def __getitem__(self, metric_name: str) -> list[np.ndarray]:
        return self._results[self._metrics.index(metric_name)]

    def __iter__(self) -> Iterator[list[np.ndarray]]:
        return iter(self._results)

    def __len__(self) -> int:
        return len(self._results)



import glob
import json
import pickle
import sys
from io import FileIO
from pathlib import Path
import biotite.structure as struc
import biotite.structure.io.mol as mol
import biotite.structure.io.pdb as pdb
import biotite.structure.io.pdbx as pdbx
import click
import pandas as pd
from peppr.evaluator import Evaluator
from peppr.metric import *
from peppr.selector import *
from peppr.version import __version__

_METRICS = {
    "monomer-rmsd": MonomerRMSD(2.0),
    "monomer-tm-score": MonomerTMScore(),
    "monomer-lddt": MonomerLDDTScore(),
    "ligand-lddt": IntraLigandLDDTScore(),
    "lddt-pli": LDDTPLIScore(),
    "lddt-ppi": LDDTPPIScore(),
    "global-lddt": GlobalLDDTScore(),
    "dockq": DockQScore(),
    "dockq-ppi": DockQScore(include_pli=False),
    "lrmsd": LigandRMSD(),
    "irmsd": InterfaceRMSD(),
    "fnat": ContactFraction(),
    "pocket-lrmsd": PocketAlignedLigandRMSD(),
    "bisy-rmsd": BiSyRMSD(2.0),
    "bond-length-violations": BondLengthViolations(),
    "clash-count": ClashCount(),
}


@click.group(
    help="It's a package for evaluation of predicted poses, right?",
    context_settings=dict(help_option_names=["-h", "--help"]),
)
@click.version_option(__version__)
def cli() -> None:
    pass


@cli.command()
@click.argument("EVALUATOR", type=click.File("wb", lazy=True))
@click.argument("METRIC", type=click.Choice(_METRICS.keys()), nargs=-1, required=True)
def create(evaluator: click.File, metric: tuple[str, ...]) -> None:
    """
    Initialize a new peppr evaluation.

    The peppr.pkl file tracking the future evaluation will be written to the given
    EVALUATOR file.
    The metrics that should be computed are given by the METRIC arguments.
    """
    metrics = [_METRICS[m] for m in metric]
    ev = Evaluator(metrics, tolerate_exceptions=True)
    _evaluator_to_file(evaluator, ev)


@cli.command()
@click.option(
    "--id",
    "-i",
    type=str,
    help="The system ID, by default taken from the reference file name",
)
@click.argument("EVALUATOR", type=click.File("rb+", lazy=True))
@click.argument(
    "REFERENCE", type=click.Path(exists=True, dir_okay=False, path_type=Path)
)
@click.argument(
    "POSE",
    type=click.Path(exists=True, dir_okay=False, path_type=Path),
    nargs=-1,
    required=True,
)
def evaluate(
    evaluator: click.File, reference: Path, pose: tuple[Path, ...], id: str | None
) -> None:
    """
    Evaluate a single system.

    Run the metrics defined by the EVALUATOR on the given system, defined by the
    REFERENCE path and POSE paths, and store the results in the EVALUATOR file.
    """
    ev = _evaluator_from_file(evaluator)
    ref = _load_system(reference)
    poses = [_load_system(path) for path in pose]
    ev.feed(id or reference.stem, ref, poses)
    _evaluator_to_file(evaluator, ev)


@cli.command()
@click.argument("EVALUATOR", type=click.File("rb+", lazy=True))
@click.argument("REFERENCE", type=str)
@click.argument("POSE", type=str)
def evaluate_batch(evaluator: click.File, reference: str, pose: str) -> None:
    """
    Evaluate multiple systems.

    Run the metrics defined by the EVALUATOR on the given systems, defined by the
    REFERENCE and POSE glob patterns, and store the results in the
    EVALUATOR file.
    If multiple poses should be evaluated for a system, POSE must be a pattern
    that matches directories of files instead of a single file.
    Note that REFERENCE and POSE are assigned to each other in
    lexicographical order.
    """
    ev = _evaluator_from_file(evaluator)
    reference_paths = sorted(
        [Path(path) for path in glob.glob(reference, recursive=True)]
    )
    pose_paths = sorted([Path(path) for path in glob.glob(pose, recursive=True)])
    if len(reference_paths) != len(pose_paths):
        raise click.UsageError(
            f"Number of reference files ({len(reference_paths)}) "
            f"does not match the number of pose files ({len(pose_paths)})"
        )

    system_ids = _find_unique_part(reference_paths)
    # Potentially remove the file suffix from the system ID
    for i, system_id in enumerate(system_ids):
        splitted = system_id.split(".")
        if len(splitted) > 1 and splitted[-1] in ["cif", "bcif", "pdb", "mol", "sdf"]:
            system_ids[i] = splitted[0]

    for system_id, ref_path, pose_path in zip(system_ids, reference_paths, pose_paths):
        if ref_path.is_dir():
            raise click.UsageError(
                "REFERENCE glob pattern must point to files, but found a directory"
            )
        reference = _load_system(ref_path)
        if pose_path.is_dir():
            poses = [
                _load_system(path)
                for path in sorted(pose_path.iterdir())
                if path.is_file()
            ]
        else:
            poses = _load_system(pose_path)
        ev.feed(system_id, reference, poses)
    _evaluator_to_file(evaluator, ev)


@cli.command()
@click.argument("EVALUATOR", type=click.File("rb", lazy=True))
@click.argument("TABLE", type=click.Path(exists=False, path_type=Path))
@click.argument("SELECTOR", type=str, nargs=-1)
def tabulate(evaluator: click.File, table: Path, selector: tuple[str, ...]) -> None:
    """
    Tabulate metric results for each system.

    Read the EVALUATOR file and write a table of the metrics for each system to the
    given TABLE CSV file.
    For systems with multiple poses, the metric results are selected by the given
    SELECTOR (may be multiple).
    Supported SELECTOR values are: 'mean', 'median', 'oracle', 'top<n>'.
    """
    ev = _evaluator_from_file(evaluator)
    df = pd.DataFrame(
        ev.tabulate_metrics(selectors=[_create_selector(sel) for sel in selector])
    )
    df.to_csv(table, index_label="System ID")


@cli.command()
@click.argument("EVALUATOR", type=click.File("rb", lazy=True))
@click.argument("SUMMARY", type=click.File("w", lazy=True))
@click.argument("SELECTOR", type=str, nargs=-1)
def summarize(
    evaluator: click.File, summary: click.File, selector: tuple[str, ...]
) -> None:
    """
    Aggregate metrics over all systems.

    Read the EVALUATOR file and write a summary of the metrics aggregated over all
    systems to the given SUMMARY .json file.
    For systems with multiple poses, the metric results are selected by the given
    SELECTOR (may be multiple).
    Supported SELECTOR values are: 'mean', 'median', 'oracle', 'top<n>'.
    """
    ev = _evaluator_from_file(evaluator)
    data = ev.summarize_metrics(selectors=[_create_selector(sel) for sel in selector])
    json.dump(data, summary, indent=2)


@cli.command()
@click.argument("METRIC", type=click.Choice(_METRICS.keys()))
@click.argument(
    "REFERENCE", type=click.Path(exists=True, dir_okay=False, path_type=Path)
)
@click.argument("POSE", type=click.Path(exists=True, dir_okay=False, path_type=Path))
def run(metric: str, reference: Path, pose: Path) -> None:
    """
    Compute a single metric for the given system.

    The given METRIC is run on the given REFERENCE and POSE and the result is
    written to STDOUT.
    """
    metric = _METRICS[metric]
    reference = _load_system(reference)
    pose = _load_system(pose)
    result = metric.evaluate(reference, pose)
    print(f"{result:.3f}", file=sys.stdout)


def _evaluator_from_file(file: FileIO) -> Evaluator:
    """
    Load a :class:`Evaluator` from the pickle representation in the given file.

    Parameters
    ----------
    file : file-like
        The file to read the pickled evaluator from.

    Returns
    -------
    Evaluator
        The evaluator.
    """
    return pickle.load(file)


def _evaluator_to_file(file: FileIO, evaluator: Evaluator) -> None:
    """
    Pickle the given :class:`Evaluator` and write it to the given file.

    Parameters
    ----------
    file : file-like
        The file to write the pickled evaluator to.
    evaluator : Evaluator
        The evaluator to pickle.
    """
    file.seek(0)
    pickle.dump(evaluator, file)


def _create_selector(selector_string: str) -> Selector:
    """
    Create a :class:`Selector` object from a string representation.

    Parameters
    ----------
    selector_string : {'mean', 'median', 'oracle', 'top<n>' and 'random<n>'}
        The string representation of the selector.

    Returns
    -------
    Selector
        The selector.
    """
    if selector_string == "mean":
        return MeanSelector()
    elif selector_string == "median":
        return MedianSelector()
    elif selector_string == "oracle":
        return OracleSelector()
    elif selector_string.startswith("top"):
        return TopSelector(int(selector_string[3:]))
    elif selector_string.startswith("random"):
        return RandomSelector(int(selector_string[6:]))
    else:
        raise click.BadParameter(f"Selector '{selector_string}' is not supported")


def _load_system(path: Path) -> struc.AtomArray:
    """
    Load a structure from a a variety of file formats.

    Parameters
    ----------
    path : Path
        The path to the structure file.
        The format is determined by the file extension.

    Returns
    -------
    AtomArray
        The system.
    """
    try:
        match path.suffix:
            case ".cif" | ".mmcif" | ".pdbx":
                cif_file = pdbx.CIFFile.read(path)
                return pdbx.get_structure(cif_file, model=1, include_bonds=True)
            case ".bcif":
                bcif_file = pdbx.BinaryCIFFile.read(path)
                return pdbx.get_structure(bcif_file, model=1, include_bonds=True)
            case ".pdb":
                pdb_file = pdb.PDBFile.read(path)
                return pdb.get_structure(pdb_file, model=1, include_bonds=True)
            case ".mol" | ".sdf":
                if path.suffix == ".sdf":
                    ctab_file = mol.SDFile.read(path)
                else:
                    ctab_file = mol.MOLFile.read(path)
                system = mol.get_structure(ctab_file)
                system.hetero[:] = True
                system.res_name[:] = "LIG"
                system.atom_name = struc.create_atom_names(system)
                return system
            case _:
                raise click.BadParameter(f"Unsupported file format '{path.suffix}'")
    except Exception as e:
        raise click.FileError(path.as_posix(), hint=str(e))


def _find_unique_part(paths: list[Path]) -> list[str]:
    """
    Find the last component of the given paths that is unique across all paths.

    Parameters
    ----------
    paths : list of Path
        The path to get the unique component from.

    Returns
    -------
    list of str
        for each path in `paths` the unique component.
    """
    # Iterate from the end to the beginning
    components = [path.parts[::-1] for path in paths]
    for i in range(max([len(c) for c in components])):
        component_of_each_path = [path_comp[i] for path_comp in components]
        if len(set(component_of_each_path)) == len(component_of_each_path):
            return component_of_each_path
    raise click.UsageError(
        "No unique system ID could be parsed from the given glob pattern"
    )



__all__ = [
    "Selector",
    "MeanSelector",
    "MedianSelector",
    "OracleSelector",
    "TopSelector",
    "RandomSelector",
]

from abc import ABC, abstractmethod
import numpy as np


class Selector(ABC):
    """
    The base class for all pose selectors.

    Its purpose is to aggregate metric values for multiple poses into a single scalar
    value.

    Attributes
    ----------
    name : str
        The name of the selector.
        Used for displaying the results via the :class:`Evaluator`.
        **ABSTRACT:** Must be overridden by subclasses.
    """

    @property
    @abstractmethod
    def name(self) -> str:
        pass

    @abstractmethod
    def select(self, values: np.ndarray, smaller_is_better: bool) -> float:
        """
        Select the *representative* metric value from a set of poses.

        The meaning of '*representative*' depends on the specific :class:`Selector`
        subclass.

        **ABSTRACT:** Must be overridden by subclasses.

        Parameters
        ----------
        values : ndarray, shape=(n,), dtype=float
            The metric values to select from.
            May contain *NaN* values.
            The values are sorted from highest to lowest confidence.
        smaller_is_better : bool
            Whether the smaller value is considered a better prediction.

        Returns
        -------
        float
            The selected value.

        Notes
        -----
        """
        raise NotImplementedError

    def __hash__(self) -> int:
        return hash(self.name)


class MeanSelector(Selector):
    """
    Selector that computes the mean of the values.
    """

    @property
    def name(self) -> str:
        return "mean"

    def select(self, values: np.ndarray, smaller_is_better: bool) -> float:
        return np.nanmean(values)


class MedianSelector(Selector):
    """
    Selector that computes the median of the values.
    """

    @property
    def name(self) -> str:
        return "median"

    def select(self, values: np.ndarray, smaller_is_better: bool) -> float:
        return np.nanmedian(values)


class OracleSelector(Selector):
    """
    Selector that returns the best value.
    """

    @property
    def name(self) -> str:
        return "Oracle"

    def select(self, values: np.ndarray, smaller_is_better: bool) -> float:
        if smaller_is_better:
            return np.nanmin(values)
        else:
            return np.nanmax(values)


class TopSelector(Selector):
    """
    Selector that returns the best value from the `k` values with highest
    confidence.

    Parameters
    ----------
    k : int
        The best value is chosen from the *k* most confident predictions.
    """

    def __init__(self, k: int) -> None:
        self._k = k

    @property
    def name(self) -> str:
        return f"Top{self._k}"

    def select(self, values: np.ndarray, smaller_is_better: bool) -> float:
        top_values = values[: self._k]
        if smaller_is_better:
            return np.nanmin(top_values)
        else:
            return np.nanmax(top_values)


class RandomSelector(Selector):
    """
    Selector that returns the best value from `k` randomly chosen values.
    Using this selector is equivalent to using the :class:`TopSelector` with
    random confidence values.

    Parameters
    ----------
    k : int
        The best value is chosen from *k* randomly chosen predictions.
    seed : int, optional
        The seed for the random number generator.
        Defaults to 42.
    """

    def __init__(self, k: int, seed: int = 42) -> None:
        self._k = k
        self._rng = np.random.default_rng(seed)

    @property
    def name(self) -> str:
        return f"Random{self._k}"

    def select(self, values: np.ndarray, smaller_is_better: bool) -> float:
        random_indices = self._rng.choice(
            range(len(values)), size=self._k, replace=False
        )
        top_values = values[random_indices]
        if smaller_is_better:
            return np.nanmin(top_values)
        else:
            return np.nanmax(top_values)



__all__ = [
    "bisy_rmsd",
]

import itertools
import biotite.structure as struc
import numpy as np
from numpy.typing import NDArray
from peppr.common import is_small_molecule
from peppr.dockq import get_contact_residues

_REPRESENTATIVE_ATOMS = ["CA", "C3'"]
_BACKBONE_ATOMS = (
    "CA",
    "C",
    "N",
    "P",
    "OP1",
    "OP2",
    "O2'",
    "O3'",
    "O4'",
    "O5'",
    "C1'",
    "C2'",
    "C3'",
    "C4'",
    "C5'",
)


def bisy_rmsd(
    reference: struc.AtomArray,
    pose: struc.AtomArray,
    inclusion_radius: float = 4.0,
    outlier_distance: float = 3.0,
    max_iterations: int = 5,
    min_anchors: int = 3,
) -> float:
    """
    Compute the *Binding-Site Superposed, Symmetry-Corrected Pose RMSD* (BiSyRMSD) for
    the given PLI complex as defined in [1]_.

    Parameters
    ----------
    reference, pose : AtomArray
        The reference and pose of the PLI complex.
    inclusion_radius : float, optional
        All residues where at least one heavy atom is within this radius of a heavy
        ligand atom, are considered part of the binding site.
        The default value is taken from [1]_.
    outlier_distance : float, optional
        The binding sites of the reference and pose are superimposed iteratively.
        In each iteration, atoms with a distance of more than this value are considered
        outliers and are removed in the next iteration.
        The default value is taken from [1]_.
        To disable outlier removal, set this value to ``inf``.
    max_iterations : int, optional
        The maximum number of iterations for the superimposition.
        The default value is taken from [1]_.
    min_anchors : int, optional
        The minimum number of anchors to use for the superimposition.
        If less than this number of anchors are present, the superimposition is
        performed on all interface backbone atoms.
        The default value is taken from [1]_.

    Returns
    -------
    float
        The *BiSyRMSD*.

    References
    ----------
    .. [1] https://doi.org/10.1002/prot.26601
    """
    receptor_mask = ~reference.hetero
    if not any(receptor_mask):
        # No receptor present
        return np.nan
    reference_ligand_chains = []
    pose_ligand_chains = []
    for start_i, stop_i in itertools.pairwise(
        struc.get_chain_starts(reference, add_exclusive_stop=True)
    ):
        reference_chain = reference[start_i:stop_i]
        if is_small_molecule(reference_chain):
            reference_ligand_chains.append(reference_chain)
            pose_ligand_chains.append(pose[start_i:stop_i])
    if len(reference_ligand_chains) == 0:
        # No ligand present
        return np.nan

    reference_ligand_coord = [chain.coord for chain in reference_ligand_chains]
    superimposed_pose_ligand_coord = [
        _superimpose_binding_site(
            reference[receptor_mask],
            pose[receptor_mask],
            reference_ligand_chain,
            inclusion_radius,
            outlier_distance,
            max_iterations,
            min_anchors,
        ).apply(pose_ligand_chain.coord)
        for reference_ligand_chain, pose_ligand_chain in zip(
            reference_ligand_chains, pose_ligand_chains
        )
    ]

    # Compute the RMSD for each ligand and take the mean
    return np.mean(
        [
            struc.rmsd(ref_coord, mod_coord).item()
            for ref_coord, mod_coord in zip(
                reference_ligand_coord, superimposed_pose_ligand_coord, strict=True
            )
        ]
    ).item()


def _superimpose_binding_site(
    reference_receptor: struc.AtomArray,
    pose_receptor: struc.AtomArray,
    reference_ligand: struc.AtomArray,
    inclusion_radius: float,
    outlier_distance: float,
    max_iterations: int,
    min_anchors: int,
) -> struc.AffineTransformation:
    """
    Get a transformation that superimposes the binding site of the pose receptor onto
    the reference receptor.

    Parameters
    ----------
    reference_receptor, pose_receptor : AtomArray
        The reference and pose receptor.
    reference_ligand : AtomArray
        The reference ligand.
        The binding site is determined by the residues in contact with the ligand atoms.
    inclusion_radius : float, optional
        All residues where at least one heavy atom is within this radius of a heavy
        ligand atom, are considered part of the binding site.
    outlier_distance : float, optional
        The binding sites of the reference and pose are superimposed iteratively.
        In each iteration, atoms with a distance of more than this value are considered
        outliers and are removed in the next iteration.
        To disable outlier removal, set this value to ``inf``.
    max_iterations : int, optional
        The maximum number of iterations for the superimposition.
    min_anchors : int, optional
        The minimum number of anchors to use for the superimposition.
        If less than this number of anchors are present, the superimposition is
        performed on all interface backbone atoms.

    Returns
    -------
    AffineTransformation
        The transformation that superimposes the binding site of the pose receptor onto
        the reference receptor.
    """
    receptor_contacts = np.unique(
        get_contact_residues(
            reference_receptor, reference_ligand, cutoff=inclusion_radius
        )[:, 0]
    )
    if len(receptor_contacts) >= min_anchors:
        interface_mask = (
            struc.get_residue_masks(reference_receptor, receptor_contacts).any(axis=0)
            # Use one atom per residue as superimposition anchor
            & np.isin(reference_receptor.atom_name, _REPRESENTATIVE_ATOMS)
        )
        transform = _superimpose_without_outliers(
            reference_receptor.coord[interface_mask],
            pose_receptor.coord[interface_mask],
            outlier_distance,
            max_iterations,
            min_anchors,
        )
    else:
        # Use all interface backbone atoms as superimposition anchor
        interface_mask = np.isin(
            reference_receptor.res_id, receptor_contacts
        ) & np.isin(reference_receptor.atom_name, _BACKBONE_ATOMS)
        _, transform = struc.superimpose(
            reference_receptor.coord[interface_mask],
            pose_receptor.coord[interface_mask],
        )
    return transform


def _superimpose_without_outliers(
    reference_coord: NDArray[np.floating],
    pose_coord: NDArray[np.floating],
    outlier_distance: float,
    max_iterations: int,
    min_anchors: int,
) -> struc.AffineTransformation:
    """
    Get a transformation that superimposes the given pose coordinates onto the
    reference coordinates.

    Outliers are iteratively removed until no outliers are left.

    Parameters
    ----------
    reference_coord, pose_coord : ndarray, shape=(n,3)
        The reference and pose coordinates.
    outlier_distance : float
        In each iteration, atoms with a distance of more than this value are considered
        outliers and are removed in the next iteration.
    max_iterations : int
        The maximum number of iterations.
    min_anchors : int
        The minimum number of anchors to use for the superimposition.
        If less than this number of anchors remain after outlier removal, the
        superimposition is performed on all interface backbone atoms.

    Returns
    -------
    biotite.structure.AffineTransformation
        The transformation that superimposes the pose onto the reference.
    """
    if max_iterations < 1:
        raise ValueError("Maximum number of iterations must be at least 1")

    # Before iterative refinement, all anchors are included
    # 'inlier' is the opposite of 'outlier'
    updated_inlier_mask = np.ones(reference_coord.shape[-2], dtype=bool)

    for _ in range(max_iterations):
        # Run superimposition
        inlier_mask = updated_inlier_mask
        filtered_reference_coord = reference_coord[..., inlier_mask, :]
        filtered_pose_coord = pose_coord[..., inlier_mask, :]
        superimposed_coord, transform = struc.superimpose(
            filtered_reference_coord, filtered_pose_coord
        )

        # Find outliers
        distance = struc.distance(filtered_reference_coord, superimposed_coord)
        updated_inlier_mask = inlier_mask.copy()
        # Distance was only calculated for the existing inliers
        # -> update the mask only for these atoms
        updated_inlier_mask[updated_inlier_mask] = distance <= outlier_distance

        if np.all(updated_inlier_mask):
            # No outliers anymore -> early termination
            break
        if np.count_nonzero(updated_inlier_mask) < min_anchors:
            # Less than min_anchors anchors would be left
            # -> revert to superimposition of all coordinates
            _, transform = struc.superimpose(reference_coord, pose_coord)
            break

    return transform



__all__ = [
    "Metric",
    "MonomerRMSD",
    "MonomerTMScore",
    "MonomerLDDTScore",
    "IntraLigandLDDTScore",
    "LDDTPLIScore",
    "LDDTPPIScore",
    "GlobalLDDTScore",
    "DockQScore",
    "LigandRMSD",
    "InterfaceRMSD",
    "ContactFraction",
    "PocketAlignedLigandRMSD",
    "BiSyRMSD",
    "BondLengthViolations",
    "ClashCount",
]

import itertools
from abc import ABC, abstractmethod
from collections import OrderedDict
from typing import Any, Callable, Dict, Tuple
import biotite.structure as struc
import numpy as np
from numpy.typing import NDArray
from peppr.bisyrmsd import bisy_rmsd
from peppr.clashes import find_clashes
from peppr.common import is_small_molecule
from peppr.dockq import (
    dockq,
    fnat,
    get_contact_residues,
    irmsd,
    lrmsd,
    pocket_aligned_lrmsd,
)


class Metric(ABC):
    """
    The base class for all evaluation metrics.

    The central :meth:`evaluate()` method takes a for a system reference and pose
    structures as input and returns a sclar score.

    Attributes
    ----------
    name : str
        The name of the metric.
        Used for displaying the results via the :class:`Evaluator`.
        **ABSTRACT:** Must be overridden by subclasses.
    thresholds : dict (str -> float)
        The named thresholds for the metric.
        Each threshold contains the lower bound
    """

    def __init__(self) -> None:
        thresholds = list(self.thresholds.values())
        if sorted(thresholds) != thresholds:
            raise ValueError("Thresholds must be sorted in ascending order")

    @property
    @abstractmethod
    def name(self) -> str:
        raise NotImplementedError

    @property
    def thresholds(self) -> OrderedDict[str, float]:
        return OrderedDict()

    @abstractmethod
    def evaluate(self, reference: struc.AtomArray, pose: struc.AtomArray) -> float:
        """
        Apply this metric on the given predicted pose with respect to the given
        reference.

        **ABSTRACT:** Must be overridden by subclasses.

        Parameters
        ----------
        reference : AtomArray, shape=(n,)
            The reference structure of the system.
            Each separate instance/molecule must have a distinct `chain_id`.
        pose : AtomArray, shape=(n,)
            The predicted pose.
            Must have the same length and atom order as the `reference`.

        Returns
        -------
        np.ndarray, shape=(m,) or None
            The metric computed for each pose.
            *NaN*, if the structure is not suitable for this metric.

        Notes
        -----
        Missing atoms in either the reference or the pose can be identified with
        *NaN* values.
        """
        raise NotImplementedError

    @abstractmethod
    def smaller_is_better(self) -> bool:
        """
        Whether as smaller value of this metric is considered a better prediction.

        **ABSTRACT:** Must be overridden by subclasses.

        Returns
        -------
        bool
            If true, a smaller value of this metric is considered a better prediction.
            Otherwise, a larger value is considered a better prediction.
        """
        raise NotImplementedError


class MonomerRMSD(Metric):
    r"""
    Compute the *root mean squared deviation* (RMSD) between each peptide chain in the
    reference and the pose and take the mean weighted by the number of heavy atoms.

    Parameters
    ----------
    threshold : float
        The RMSD threshold to use for the *good* predictions.
    ca_only : bool, optional
        If ``True``, only consider :math:`C_{\alpha}` atoms.
        Otherwise, consider all heavy atoms.
    """

    def __init__(self, threshold: float, ca_only: bool = True) -> None:
        self._threshold = threshold
        self._ca_only = ca_only
        super().__init__()

    @property
    def name(self) -> str:
        if self._ca_only:
            return "CA-RMSD"
        else:
            return "all-atom RMSD"

    @property
    def thresholds(self) -> OrderedDict[str, float]:
        return OrderedDict(
            [(f"<{self._threshold}", 0), (f">{self._threshold}", self._threshold)]
        )

    def evaluate(self, reference: struc.AtomArray, pose: struc.AtomArray) -> float:
        def superimpose_and_rmsd(reference_chain, pose_chain):  # type: ignore[no-untyped-def]
            pose_chain, _ = struc.superimpose(reference_chain, pose_chain)
            return struc.rmsd(reference_chain, pose_chain)

        mask = ~reference.hetero
        if self._ca_only:
            mask &= reference.atom_name == "CA"
        return _run_for_each_monomer(reference[mask], pose[mask], superimpose_and_rmsd)

    def smaller_is_better(self) -> bool:
        return True


class MonomerTMScore(Metric):
    """
    Compute the *TM-score* score for each monomer and take the mean weighted by the
    number of atoms.
    """

    @property
    def name(self) -> str:
        return "TM-score"

    def evaluate(self, reference: struc.AtomArray, pose: struc.AtomArray) -> float:
        def superimpose_and_tm_score(reference, pose):  # type: ignore[no-untyped-def]
            # Use 'superimpose_structural_homologs()' instead of 'superimpose()',
            # as it optimizes the TM-score instead of the RMSD
            try:
                super, _, ref_i, pose_i = struc.superimpose_structural_homologs(
                    reference, pose, max_iterations=1
                )
                return struc.tm_score(reference, super, ref_i, pose_i)
            except ValueError as e:
                if "No anchors found" in str(e):
                    # The structures are too dissimilar for structure-based
                    # superimposition, i.e. the pose is very bad
                    return 0.0
                else:
                    raise

        # TM-score is only defined for peptide chains
        mask = struc.filter_amino_acids(reference) & ~reference.hetero
        reference = reference[mask]
        pose = pose[mask]
        return _run_for_each_monomer(reference, pose, superimpose_and_tm_score)

    def smaller_is_better(self) -> bool:
        return False


class MonomerLDDTScore(Metric):
    """
    Compute the *local Distance Difference Test* (lDDT) score for each monomer
    and take the mean weighted by the number of atoms.
    """

    @property
    def name(self) -> str:
        return "intra protein lDDT"

    def evaluate(self, reference: struc.AtomArray, pose: struc.AtomArray) -> float:
        protein_mask = ~reference.hetero
        if not protein_mask.any():
            # No protein present
            return np.nan
        reference = reference[protein_mask]
        pose = pose[protein_mask]
        return _run_for_each_monomer(reference, pose, struc.lddt)

    def smaller_is_better(self) -> bool:
        return False


class IntraLigandLDDTScore(Metric):
    """
    Compute the *local Distance Difference Test* (lDDT) score for contacts within each
    small molecule.
    """

    @property
    def name(self) -> str:
        return "intra ligand lDDT"

    def evaluate(self, reference: struc.AtomArray, pose: struc.AtomArray) -> float:
        def within_same_molecule(contacts: np.ndarray) -> np.ndarray:
            # Find the index of the chain/molecule for each atom
            chain_indices = struc.get_chain_positions(
                reference, contacts.flatten()
            ).reshape(contacts.shape)
            # Remove contacts between atoms of different molecules
            return chain_indices[:, 0] == chain_indices[:, 1]

        ligand_mask = reference.hetero
        if not ligand_mask.any():
            # No ligands present
            return np.nan
        reference = reference[ligand_mask]
        pose = pose[ligand_mask]
        return struc.lddt(
            reference,
            pose,
            exclude_same_residue=False,
            filter_function=within_same_molecule,
        ).item()

    def smaller_is_better(self) -> bool:
        return False


class LDDTPLIScore(Metric):
    """
    Compute the CASP LDDT-PLI score, i.e. the lDDT for protein-ligand interactions
    as defined by [1]_.

    References
    ----------
    .. [1] https://doi.org/10.1002/prot.26601
    """

    @property
    def name(self) -> str:
        return "protein-ligand lDDT"

    def evaluate(self, reference: struc.AtomArray, pose: struc.AtomArray) -> float:
        def lddt_pli(reference, pose):  # type: ignore[no-untyped-def]
            ligand_mask = reference.hetero
            polymer_mask = ~ligand_mask
            binding_site_contacts = np.unique(
                get_contact_residues(
                    reference[polymer_mask], reference[ligand_mask], cutoff=4.0
                )[:, 0]
            )
            # No contacts between the ligand and protein in reference -> no binding site
            # -> metric is undefined
            if len(binding_site_contacts) == 0:
                return np.nan
            binding_site_mask = struc.get_residue_masks(
                reference, binding_site_contacts
            ).any(axis=0)
            return struc.lddt(
                reference,
                pose,
                atom_mask=ligand_mask,
                partner_mask=binding_site_mask,
                inclusion_radius=6.0,
                distance_bins=(0.5, 1.0, 2.0, 4.0),
                symmetric=True,
            )

        return _average_over_ligands(reference, pose, lddt_pli)

    def smaller_is_better(self) -> bool:
        return False


class LDDTPPIScore(Metric):
    """
    Compute the the lDDT for protein-protein interactions, i.e. all intra-chain
    contacts are not included.
    """

    @property
    def name(self) -> str:
        return "protein-protein lDDT"

    def evaluate(self, reference: struc.AtomArray, pose: struc.AtomArray) -> float:
        protein_mask = ~reference.hetero
        if not protein_mask.any():
            # This is not a PPI system
            return np.nan
        reference = reference[protein_mask]
        pose = pose[protein_mask]
        return struc.lddt(reference, pose, exclude_same_chain=True)

    def smaller_is_better(self) -> bool:
        return False


class GlobalLDDTScore(Metric):
    r"""
    Compute the lDDT score for all contacts in the system, i.e. both intra- and
    inter-chain contacts.
    This is equivalent to the original lDDT definition in [1]_.

    Parameters
    ----------
    backbone_only : bool, optional
        If ``True``, only consider :math:`C_{\alpha}` from peptides and :math:`C_3^'`
        from nucleic acids.
        Otherwise, consider all heavy atoms.

    References
    ----------
    .. [1] https://doi.org/10.1093/bioinformatics/btt473
    """

    def __init__(self, backbone_only: bool = True) -> None:
        self._backbone_only = backbone_only
        super().__init__()

    @property
    def name(self) -> str:
        if self._backbone_only:
            return "global backbone lDDT"
        else:
            return "global all-atom lDDT"

    def evaluate(self, reference: struc.AtomArray, pose: struc.AtomArray) -> float:
        if self._backbone_only:
            mask = ~reference.hetero & np.isin(reference.atom_name, ["CA", "C3'"])
            reference = reference[mask]
            pose = pose[mask]
        if reference.array_length() == 0:
            return np.nan
        return struc.lddt(reference, pose).item()

    def smaller_is_better(self) -> bool:
        return False


class DockQScore(Metric):
    """
    Compute the *DockQ* score for the given complex as defined in [1]_.

    Parameters
    ----------
    include_pli : bool, optional
        If set to ``False``, small molecules are excluded from the calculation.

    References
    ----------
    .. [1] https://doi.org/10.1093/bioinformatics/btae586
    """

    def __init__(self, include_pli: bool = True) -> None:
        self._include_pli = include_pli
        super().__init__()

    @property
    def name(self) -> str:
        if self._include_pli:
            return "DockQ"
        else:
            return "DockQ-PPI"

    @property
    def thresholds(self) -> OrderedDict[str, float]:
        return OrderedDict(
            [
                ("incorrect", 0.0),
                ("acceptable", 0.23),
                ("medium", 0.49),
                ("high", 0.80),
            ]
        )

    def evaluate(self, reference: struc.AtomArray, pose: struc.AtomArray) -> float:
        def run_dockq(reference_chain1, reference_chain2, pose_chain1, pose_chain2):  # type: ignore[no-untyped-def]
            if not self._include_pli and (
                is_small_molecule(reference_chain1)
                or is_small_molecule(reference_chain2)
            ):
                # Do not compute DockQ for PLI pairs if disabled
                return np.nan
            if is_small_molecule(reference_chain1) and is_small_molecule(
                reference_chain2
            ):
                # Do not compute DockQ for small molecule pairs
                return np.nan
            return dockq(
                *_select_receptor_and_ligand(
                    reference_chain1, reference_chain2, pose_chain1, pose_chain2
                )
            ).score

        return _run_for_each_chain_pair(reference, pose, run_dockq)

    def smaller_is_better(self) -> bool:
        return False


class LigandRMSD(Metric):
    """
    Compute the *Ligand RMSD* for the given protein complex as defined in [1]_.
    The score is first separately computed for all pairs of chains that are in contact,
    and the averaged. If the reference doesn't contain any chains in contact, *NaN* is returned.

    References
    ----------
    .. [1] https://doi.org/10.1093/bioinformatics/btae586
    """

    @property
    def name(self) -> str:
        return "LRMSD"

    def evaluate(self, reference: struc.AtomArray, pose: struc.AtomArray) -> float:
        mask = struc.filter_amino_acids(reference) & ~reference.hetero
        reference = reference[mask]
        pose = pose[mask]

        def lrmsd_on_interfaces_only(
            reference_chain1: struc.AtomArray,
            reference_chain2: struc.AtomArray,
            pose_chain1: struc.AtomArray,
            pose_chain2: struc.AtomArray,
        ) -> float | np.floating | NDArray[np.floating]:
            reference_contacts = get_contact_residues(
                reference_chain1,
                reference_chain2,
                cutoff=10.0,
            )

            if len(reference_contacts) == 0:
                return np.nan
            else:
                return lrmsd(
                    *_select_receptor_and_ligand(
                        reference_chain1, reference_chain2, pose_chain1, pose_chain2
                    )
                )

        return _run_for_each_chain_pair(reference, pose, lrmsd_on_interfaces_only)

    def smaller_is_better(self) -> bool:
        return True


class InterfaceRMSD(Metric):
    """
    Compute the *Interface RMSD* for the given protein complex as defined in [1]_.

    References
    ----------
    .. [1] https://doi.org/10.1093/bioinformatics/btae586
    """

    @property
    def name(self) -> str:
        return "iRMSD"

    def evaluate(self, reference: struc.AtomArray, pose: struc.AtomArray) -> float:
        mask = struc.filter_amino_acids(reference) & ~reference.hetero
        reference = reference[mask]
        pose = pose[mask]
        # iRMSD is independent of the selection of receptor and ligand chain
        return _run_for_each_chain_pair(reference, pose, irmsd)

    def smaller_is_better(self) -> bool:
        return True


class ContactFraction(Metric):
    """
    Compute the fraction of correctly predicted reference contacts (*Fnat*) as defined
    in [1]_.

    References
    ----------
    .. [1] https://doi.org/10.1093/bioinformatics/btae586
    """

    @property
    def name(self) -> str:
        return "fnat"

    def evaluate(self, reference: struc.AtomArray, pose: struc.AtomArray) -> float:
        mask = struc.filter_amino_acids(reference) & ~reference.hetero
        reference = reference[mask]
        pose = pose[mask]
        # Fnat is independent of the selection of receptor and ligand chain
        # Caution: `fnat()` returns both fnat and fnonnat -> select first element
        return _run_for_each_chain_pair(reference, pose, lambda *args: fnat(*args)[0])

    def smaller_is_better(self) -> bool:
        return False


class PocketAlignedLigandRMSD(Metric):
    """
    Compute the *Pocket aligned ligand RMSD* for the given PLI complex as defined
    in [1]_.

    References
    ----------
    .. [1] https://doi.org/10.1093/bioinformatics/btae586
    """

    @property
    def name(self) -> str:
        return "PLI-LRMSD"

    def evaluate(self, reference: struc.AtomArray, pose: struc.AtomArray) -> float:
        def run_lrmd(reference_chain1, reference_chain2, pose_chain1, pose_chain2):  # type: ignore[no-untyped-def]
            n_small_molecules = sum(
                is_small_molecule(chain)
                for chain in [reference_chain1, reference_chain2]
            )
            if n_small_molecules != 1:
                # Either two proteins or two small molecules -> not a valid PLI pair
                return np.nan
            return pocket_aligned_lrmsd(
                *_select_receptor_and_ligand(
                    reference_chain1, reference_chain2, pose_chain1, pose_chain2
                )
            )

        return _run_for_each_chain_pair(reference, pose, run_lrmd)

    def smaller_is_better(self) -> bool:
        return True


class BiSyRMSD(Metric):
    """
    Compute the *Binding-Site Superposed, Symmetry-Corrected Pose RMSD* (BiSyRMSD) for
    the given PLI complex.

    The method and default parameters are described in [1]_.

    Parameters
    ----------
    threshold : float
        The RMSD threshold to use for the *good* predictions.
    inclusion_radius : float, optional
        All residues where at least one heavy atom is within this radius of a heavy
        ligand atom, are considered part of the binding site.
    outlier_distance : float, optional
        The binding sites of the reference and pose are superimposed iteratively.
        In each iteration, atoms with a distance of more than this value are considered
        outliers and are removed in the next iteration.
        To disable outlier removal, set this value to ``inf``.
    max_iterations : int, optional
        The maximum number of iterations for the superimposition.
    min_anchors : int, optional
        The minimum number of anchors to use for the superimposition.
        If less than this number of anchors are present, the superimposition is
        performed on all interface backbone atoms.

    References
    ----------
    .. [1] https://doi.org/10.1002/prot.26601
    """

    def __init__(
        self,
        threshold: float,
        inclusion_radius: float = 4.0,
        outlier_distance: float = 3.0,
        max_iterations: int = 5,
        min_anchors: int = 3,
    ) -> None:
        self._threshold = threshold
        self._inclusion_radius = inclusion_radius
        self._outlier_distance = outlier_distance
        self._max_iterations = max_iterations
        self._min_anchors = min_anchors
        super().__init__()

    @property
    def name(self) -> str:
        return "BiSyRMSD"

    @property
    def thresholds(self) -> OrderedDict[str, float]:
        return OrderedDict(
            [(f"<{self._threshold}", 0), (f">{self._threshold}", self._threshold)]
        )

    def evaluate(self, reference: struc.AtomArray, pose: struc.AtomArray) -> float:
        return bisy_rmsd(
            reference,
            pose,
            inclusion_radius=self._inclusion_radius,
            outlier_distance=self._outlier_distance,
            max_iterations=self._max_iterations,
            min_anchors=self._min_anchors,
        )

    def smaller_is_better(self) -> bool:
        return True


class BondLengthViolations(Metric):
    """
    Check for unusual bond lengths in the structure by comparing against reference values.
    Returns the percentage of bonds that are within acceptable ranges.

    Parameters
    ----------
    tolerance : float, optional
        The tolerance in Angstroms for acceptable deviation from ideal bond lengths.
        Default is 0.1 Angstroms.
    reference_bonds : dict, optional
        Dictionary mapping atom type pairs to ideal bond lengths.
        If not provided, uses a default set of common bond lengths.
    """

    # Default reference bond lengths in Angstroms
    _DEFAULT_REFERENCE_BONDS = {
        ("C", "C"): np.array([1.54, 1.33, 1.20]),
        ("C", "N"): np.array([1.47, 1.27, 1.15]),
        ("C", "O"): np.array([1.43, 1.21]),
        ("N", "H"): np.array([1.01]),
        ("O", "H"): np.array([0.96]),
        ("C", "H"): np.array([1.09]),
        ("C", "S"): np.array([1.82]),
        ("S", "S"): np.array([2.05]),
    }

    def __init__(
        self,
        tolerance: float = 0.1,
        reference_bonds: Dict[Tuple[str, str], np.ndarray] | None = None,
    ) -> None:
        self._tolerance = tolerance
        self._reference_bonds = (
            reference_bonds
            if reference_bonds is not None
            else self._DEFAULT_REFERENCE_BONDS
        )
        # Add reverse pairs for convenience
        reverse_bonds = {(b, a): v for (a, b), v in self._reference_bonds.items()}
        self._reference_bonds.update(reverse_bonds)
        super().__init__()

    @property
    def name(self) -> str:
        return "Bond-length-violation"

    @property
    def thresholds(self) -> OrderedDict[str, float]:
        return OrderedDict(
            [
                ("poor", 0.0),
                ("acceptable", 0.9),
                ("good", 0.95),
                ("excellent", 0.99),
            ]
        )

    def evaluate(self, reference: struc.AtomArray, pose: struc.AtomArray) -> float:
        """
        Calculate the percentage of bonds that are outside acceptable ranges.

        Parameters
        ----------
        reference : AtomArray
            Not used in this metric as we compare against ideal bond lengths.
        pose : AtomArray
            The structure to evaluate.

        Returns
        -------
        float
            Percentage of bonds outside acceptable ranges (0.0 to 1.0).
        """
        if pose.bonds is None:
            return np.nan

        total_checked = 0
        valid_bonds = 0

        # Get all bonds and iterate through the bond tuples
        for i, j, _ in pose.bonds.as_array():
            atom1_type = pose.element[i]
            atom2_type = pose.element[j]

            if (atom1_type, atom2_type) in self._reference_bonds:
                total_checked += 1
                bond_length = struc.distance(pose[i], pose[j])
                ideal_lengths = self._reference_bonds[(atom1_type, atom2_type)]
                if np.any(np.abs(bond_length - ideal_lengths) <= self._tolerance):
                    valid_bonds += 1

        if total_checked == 0:
            return np.nan

        return 1 - (valid_bonds / total_checked)

    def smaller_is_better(self) -> bool:
        return True


class ClashCount(Metric):
    """
    Count the number of clashes between atoms in the pose.
    """

    @property
    def name(self) -> str:
        return "Number of clashes"

    def evaluate(self, reference: struc.AtomArray, pose: struc.AtomArray) -> float:
        if pose.array_length() == 0:
            return np.nan
        return len(find_clashes(pose))

    def smaller_is_better(self) -> bool:
        return True


def _run_for_each_monomer(
    reference: struc.AtomArray,
    pose: struc.AtomArray,
    function: Callable[[struc.AtomArray, struc.AtomArray], float],
) -> float:
    """
    Run the given function for each monomer in the reference and pose.

    Parameters
    ----------
    reference : AtomArray, shape=(n,)
        The reference structure of the system.
    pose : AtomArray, shape=(n,)
        The predicted pose.
        Must have the same length and atom order as the `reference`.
    function : Callable
        The function to run for each monomer.
        Takes the reference and pose as input and returns a scalar value.

    Returns
    -------
    metrics : float
        The average return value of `function`, weighted by the number of atoms.
        If the input structure contains no chains, *NaN* is returned.
    """
    values = []
    chain_starts = struc.get_chain_starts(reference, add_exclusive_stop=True)
    for start_i, stop_i in itertools.pairwise(chain_starts):
        reference_chain = reference[start_i:stop_i]
        pose_chain = pose[start_i:stop_i]
        values.append(function(reference_chain, pose_chain))
    values = np.array(values)

    if len(values) == 0:
        # No chains in the structure
        return np.nan
    else:
        n_atoms_per_chain = np.diff(chain_starts)
        # Ignore chains where the values are NaN
        not_nan_mask = np.isfinite(values)
        return np.average(
            values[not_nan_mask], weights=n_atoms_per_chain[not_nan_mask]
        ).item()


def _run_for_each_chain_pair(
    reference: struc.AtomArray,
    pose: struc.AtomArray,
    function: Callable[
        [struc.AtomArray, struc.AtomArray, struc.AtomArray, struc.AtomArray], float
    ],
) -> float:
    """
    Run the given function for each chain pair combination in the reference and pose
    and return the average value.

    Parameters
    ----------
    reference : AtomArray, shape=(n,)
        The reference structure of the system.
    pose : AtomArray, shape=(n,)
        The predicted pose.
        Must have the same length and atom order as the `reference`.
    function : Callable[(reference_chain1, reference_chain2, pose_chain1, pose_chain2), float]
        The function to run for each interface.
        Takes the reference and pose chains in contact as input and returns a scalar
        value.
        The function may also return *NaN*, if its result should be ignored.

    Returns
    -------
    metrics : float
        The average return value of `function`, weighted by the number of atoms.
        If the input structure contains only one chain, *NaN* is returned.
    """
    chain_starts = struc.get_chain_starts(reference, add_exclusive_stop=True)
    reference_chains = [
        reference[start:stop] for start, stop in itertools.pairwise(chain_starts)
    ]
    pose_chains = [pose[start:stop] for start, stop in itertools.pairwise(chain_starts)]
    results = []
    for chain_i, chain_j in itertools.combinations(range(len(reference_chains)), 2):
        results.append(
            function(
                reference_chains[chain_i],
                reference_chains[chain_j],
                pose_chains[chain_i],
                pose_chains[chain_j],
            )
        )
    if len(results) == 0:
        return np.nan
    return np.nanmean(results).item()


def _average_over_ligands(
    reference: struc.AtomArray,
    pose: struc.AtomArray,
    function: Callable[[struc.AtomArray, struc.AtomArray], float],
) -> float:
    """
    Run the given function for each ligand in the reference and pose.
    and return the average value.

    Parameters
    ----------
    reference : AtomArray, shape=(n,)
        The reference structure of the system.
    pose : AtomArray, shape=(n,)
        The predicted pose.
        Must have the same length and atom order as the `reference`.
    function : Callable[[reference_ligand, pose_ligand], float]
        The function to run for each ligand.
        Takes the reference and pose ligand as input and returns a scalar
        value.

    Returns
    -------
    metrics : float
        The average return value of `function`, weighted by the number of atoms.
        If the input structure contains no ligand atoms, *NaN* is returned.
    """
    values = _run_for_each_ligand(reference, pose, function)
    if len(values) == 0:
        # No ligands in the structure
        return np.nan
    else:
        return np.mean(values).item()


def _run_for_each_ligand(
    reference: struc.AtomArray,
    pose: struc.AtomArray,
    function: Callable[[struc.AtomArray, struc.AtomArray], Any],
) -> list[Any]:
    """
    Run the given function for each isolated ligand in complex with all proteins from
    the system.

    Parameters
    ----------
    reference : AtomArray, shape=(n,)
        The reference structure of the system.
    pose : AtomArray, shape=(n,)
        The predicted pose.
        Must have the same length and atom order as the `reference`.
    function : Callable[[reference_ligand, pose_ligand], float]
        The function to run.
        Takes the reference and pose system as input and returns a scalar
        value.

    Returns
    -------
    metrics : float
        The average return value of `function`.
        If the input structure contains no ligand atoms, *NaN* is returned.
    """
    values = []
    ligand_mask = reference.hetero
    polymer_mask = ~ligand_mask
    chain_starts = struc.get_chain_starts(reference)
    if len(chain_starts) == 0:
        # No chains in the structure
        return []
    chain_masks = struc.get_chain_masks(reference, chain_starts)
    # Only keep chain masks that correspond to ligand chains
    ligand_masks = chain_masks[(chain_masks & ligand_mask).any(axis=-1)]
    for ligand_mask in ligand_masks:
        # Evaluate each isolated ligand in complex separately
        complex_mask = ligand_mask | polymer_mask
        values.append(function(reference[complex_mask], pose[complex_mask]))
    return values


def _select_receptor_and_ligand(
    reference_chain1: struc.AtomArray,
    reference_chain2: struc.AtomArray,
    pose_chain1: struc.AtomArray,
    pose_chain2: struc.AtomArray,
) -> tuple[struc.AtomArray, struc.AtomArray, struc.AtomArray, struc.AtomArray]:
    """
    Select the receptor and ligand for the given interface.

    The longer of both chains is the receptor.

    Parameters
    ----------
    reference_chain1, reference_chain2 : AtomArray, shape=(n,)
        The reference chains.
    pose_chain1, pose_chain2 : AtomArray, shape=(n,)
        The pose chains.

    Returns
    -------
    reference_receptor, reference_ligand, pose_receptor, pose_ligand : AtomArray
        The selected receptor and ligand.
    """
    if is_small_molecule(reference_chain1):
        return (reference_chain2, reference_chain1, pose_chain2, pose_chain1)
    elif is_small_molecule(reference_chain2):
        return (reference_chain1, reference_chain2, pose_chain1, pose_chain2)
    elif reference_chain1.array_length() >= reference_chain2.array_length():
        return (reference_chain1, reference_chain2, pose_chain1, pose_chain2)
    else:
        return (reference_chain2, reference_chain1, pose_chain2, pose_chain1)



# mypy: disable-error-code="attr-defined"

__all__ = ["sanitize"]

import rdkit
import rdkit.Chem.AllChem as Chem


def sanitize(mol: Chem.Mol, max_fix_iterations: int = 1000) -> None:
    """
    Fix small issues with RDKit SanitizeMol and sanitize molecule.

    This is an alternative to using :func:`SanitizeMol()` directly, in cases it fails
    due to issues with the molecule which are fixed by this function.

    Parameters
    ----------
    mol : Chem.rdchem.Mol
        The molecule to sanitize.
    max_fix_iterations : int, optional
        The maximum number of iterations to fix problems.

    Notes
    -----
    Deals with cases:

    - ``AtomValenceException``: add charge if total valence is exceeding default (for N, O atoms only)
    - ``KekulizeException``: for ring N atoms with unspecified protonation
    """
    mol.UpdatePropertyCache(strict=False)
    prev_problems: list[Exception] = []

    # RDKit detects probels iteratively,
    # i.e. new problems may show up as soon as previous problems are fixed
    for _ in range(max_fix_iterations):
        with rdkit.rdBase.BlockLogs():
            # Temporarily disable RDKit verbosity while in scope
            problems = Chem.DetectChemistryProblems(mol)
        if not problems:
            # All problematic places have been fixed
            break
        for problem in problems:
            for prev_problem in prev_problems:
                if _is_same_problem(problem, prev_problem):
                    # Fixing this problem previously failed,
                    # so there is no need to try again
                    break
        for problem in problems:
            _fix_problem(mol, problem)
        prev_problems = list(problems)

    with rdkit.rdBase.BlockLogs():
        # if any issue remains - it will be caught as an Error
        Chem.SanitizeMol(mol)


def _fix_problem(mol: Chem.Mol, problem: Exception) -> None:
    """
    Apply fixes for common problems.

    Parameters
    ----------
    mol : Chem.rdchem.Mol
        The molecule to fix.
    problem : Exception
        The problem to fix.
    """
    pt = Chem.GetPeriodicTable()
    if problem.GetType() == "AtomValenceException":
        at = mol.GetAtomWithIdx(problem.GetAtomIdx())
        elem = at.GetAtomicNum()
        if elem == 6:
            # we do not like charged carbons, thus pass without modification
            return
        elif elem in [7, 8]:
            # only process positively charged N, O atoms
            opt_charge = at.GetExplicitValence() - pt.GetDefaultValence(elem)
            formal_charge = at.GetFormalCharge()
            if opt_charge > formal_charge:
                if abs(opt_charge) > 1:
                    raise AssertionError(f"absolute expected charge > 1: {opt_charge}")
                else:
                    # fix explicit valence issue
                    at.SetFormalCharge(opt_charge)
        else:
            # for higher elements like S - may need to use pt.GetValenceList()
            # currently not implemented
            return
    if problem.GetType() == "KekulizeException":
        # hack: only works for nitrogens with missing explicit Hs
        for atidx in problem.GetAtomIndices():
            at = mol.GetAtomWithIdx(atidx)
            # set one of the nitrogens with two bonds in a ring system as "[nH]"
            if at.GetAtomicNum() == 7 and at.GetDegree() == 2:
                at.SetNumExplicitHs(1)
                break


def _is_same_problem(problem1: Exception, problem2: Exception) -> bool:
    """
    Check if two exceptions related to chemistry problems are the same.

    Parameters
    ----------
    problem1, problem2 : Exception
        The problems to compare.

    Returns
    -------
    same : bool
        True if the two problems are the same.
    """
    if problem1.GetType() != problem2.GetType():
        return False
    if problem1.GetAtomIndices() != problem2.GetAtomIndices():
        return False
    return True



__all__ = ["find_clashes"]

import biotite.structure as struc
import biotite.structure.info as info
import numpy as np
from numpy.typing import NDArray


def find_clashes(atoms: struc.AtomArray, vdw_scaling: float = 0.65) -> NDArray[np.int_]:
    """
    Find atom clashes in the given structure.

    A clash is defined as a pair of non-bonded atoms whose distance is smaller than
    the sum of their Van-der-Waals radii (multiplied by a scaling factor).

    Parameters
    ----------
    atoms : AtomArray
        The structure to find the clashes in.
    vdw_scaling : float, optional
        The scaling factor for the Van-der-Waals radii.

    Returns
    -------
    ndarray, shape=(n,2), dtype=int
        The array of clashes.
        Each element represents a pair of atom indices that are in clash.
    """
    if atoms.array_length() == 0:
        return np.zeros((0, 2), dtype=int)

    # Although we only consider heavy atoms,
    # we cannot properly use ProtOr radii as they lead to intra residue clashes
    vdw_radii = np.array([info.vdw_radius_single(e) for e in atoms.element])

    # Any pair of atoms, whose distance is larger than the largest possible VdW radius,
    # is not relevant for clash detection
    max_clash_threshold = 2 * vdw_scaling * np.max(vdw_radii)
    contacts = _find_contacts(atoms, max_clash_threshold)
    distances = struc.index_distance(atoms, contacts)
    clash_thresholds = vdw_scaling * (
        vdw_radii[contacts[:, 0]] + vdw_radii[contacts[:, 1]]
    )
    clashes = contacts[distances < clash_thresholds]

    # Atoms that are connected by a bond are not considered clashes
    bond_set = set([frozenset((i, j)) for i, j, _ in atoms.bonds.as_array()])
    clash_set = set([frozenset((i, j)) for i, j in clashes])
    clash_set -= bond_set
    # Ensure two dimensional array, even if there are no clashes
    return np.array([tuple(e) for e in clash_set], dtype=int).reshape(-1, 2)


def _find_contacts(
    atoms: struc.AtomArray,
    inclusion_radius: float,
) -> NDArray[np.int_]:
    """
    Find contacts between the atoms in the given structure.

    Parameters
    ----------
    atoms : AtomArray
        The structure to find the contacts for.
    inclusion_radius : float
        Pairwise atom distances are considered within this radius.

    Returns
    -------
    ndarray, shape=(n,2), dtype=int
        The array of contacts.
        Each element represents a pair of atom indices that are in contact.
    """
    coords = struc.coord(atoms)
    # Use a cell list to find atoms within inclusion radius in O(n) time complexity
    cell_list = struc.CellList(coords, inclusion_radius)
    # Pairs of indices for atoms within the inclusion radius
    all_contacts = cell_list.get_atoms(coords, inclusion_radius)
    # Convert into pairs of indices
    contacts = _to_sparse_indices(all_contacts)
    # Do not consider self-contacts
    contacts = contacts[contacts[:, 0] != contacts[:, 1]]
    return contacts


def _to_sparse_indices(all_contacts: NDArray[np.int_]) -> NDArray[np.int_]:
    """
    Create tuples of contact indices from the :meth:`CellList.get_atoms()` return value.

    In other words, they would mark the non-zero elements in a dense contact matrix.

    Parameters
    ----------
    all_contacts : ndarray, dtype=int, shape=(m,n)
        The contact indices as returned by :meth:`CellList.get_atoms()`.
        Padded with -1, in the second dimension.
        Dimension *m* marks the query atoms, dimension *n* marks the contact atoms.

    Returns
    -------
    ndarray, dtype=int, shape=(l,2)
        The contact indices.
        Each column contains the query and contact atom index.
    """
    # Find rows where a query atom has at least one contact
    non_empty_indices = np.where(np.any(all_contacts != -1, axis=1))[0]
    # Take those rows and flatten them
    contact_indices = all_contacts[non_empty_indices].flatten()
    # For each row the corresponding query atom is the same
    # Hence in the flattened form the query atom index is simply repeated
    query_indices = np.repeat(non_empty_indices, all_contacts.shape[1])
    combined_indices = np.stack([query_indices, contact_indices], axis=1)
    # Remove the padding values
    return combined_indices[contact_indices != -1]



from .bisyrmsd import *
from .clashes import *
from .common import *
from .dockq import *
from .evaluator import *
from .match import *
from .metric import *
from .sanitize import *
from .selector import *
from .version import *
